# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals
import os
import sys

from byoc import (
    stacks,
    stores,
)

# python2 is not supported
if sys.version_info >= (3,):
    import importlib.util as imputil


class StartingNameMatcher(stacks.SectionMatcher):
    """A sub name section matcher.

    This selects sections starting with a given name with sections sorted
    alphanumerically.
    """

    def __init__(self, store, name):
        super(StartingNameMatcher, self).__init__(store)
        self.name = name

    def get_sections(self):
        """Get all sections starting with ``name`` in the store.

        The most generic sections are described first in the store, then more
        specific ones can be provided for reduced scopes.

        The returned sections are therefore returned in the reversed order so
        the most specific ones can be found first.
        """
        store = self.store
        # Longer sections are more specific, they should be returned first
        for _, section in sorted(
                store.get_sections(), reverse=True,
                # s[1].id or '' is to cope with None as a section id
                key=lambda s: s[1].id or ''):
            if section.id is None:
                # The no-name section is always included if present
                yield store, section
                continue
            if self.name is None:
                # If no name has been defined, no section can match
                continue
            section_name = section.id
            if self.name.startswith(section_name):
                yield store, section


VmMatcher = StartingNameMatcher
VmStore = stores.FileStore
VmCmdLineStore = stores.CommandLineStore


def system_config_dir():
    return '/etc/byov'


def user_config_dir():
    return os.path.expanduser('~/.config/byov')


def config_file_basename():
    return 'byov.conf'


def existing_config_file_basename():
    return 'existing-vms.conf'


def config_files_in(directory, conf_dir='conf.d', suffix='.conf'):
    """Iterate valid config file names in a directory."""
    fulldir = os.path.join(directory, conf_dir)
    if os.path.exists(fulldir):
        for p in sorted(os.listdir(fulldir)):
            # Filter out if not ending with suffix (so README files, for
            # example could be added or renaming can be used to disable config
            # files)
            if p.endswith(suffix):
                yield os.path.join(fulldir, p)


class VmStack(stacks.Stack):
    """Per-vm options."""

    def __init__(self, name):
        """Make a new stack for a given vm.

        :param name: The name of a virtual machine.

        The options are searched in following files for all sections matching
        ``name``. Additionally, defaults can be provided in the no-name
        section, except for ``existing-vms.conf`` which uses only the ``name``
        section.

        * the directory local file: ./byov.conf
        * the directory files in alphanumeric order: byov.conf.d/*.conf.
        * the user file: ~/.config/byov/byov.conf
        * the user files in alphanumeric order: ~/.config/byov/conf.d/*.conf.
        * the system-wide file: /etc/byov.conf
        * the system-wide files in alphanumeric order: /etc/byov/conf.d/*.conf.
        * the existing vms file (for vms already setup):
          ~/.config/byov/existing-vms.conf

        Longer section names are more specific than shorter ones across all
        files.
        """
        self.vm_name = name
        # FIXME: Would life be simpler if the cmdline store was shared ?  Or
        # not ? -- vila 2018-02-12

        # FIXME: Hard-coding a class name is a smell, at the very least make
        # this a class attribute. -- vila 2019-08-13
        self.cmdline_store = VmCmdLineStore()

        section_getters = []

        lpath = os.path.abspath(config_file_basename())
        self.local_store = self.get_shared_store(VmStore(lpath))
        section_getters.append(VmMatcher(self.local_store, name).get_sections)

        for cpath in config_files_in(os.path.abspath('.'), 'byov.conf.d'):
            store = self.get_shared_store(VmStore(cpath))
            section_getters.append(VmMatcher(store, name).get_sections)

        upath = os.path.join(user_config_dir(), config_file_basename())
        self.user_store = self.get_shared_store(VmStore(upath))
        section_getters.append(VmMatcher(self.user_store, name).get_sections)

        for cpath in config_files_in(user_config_dir()):
            store = self.get_shared_store(VmStore(cpath))
            section_getters.append(VmMatcher(store, name).get_sections)

        spath = os.path.join(system_config_dir(), config_file_basename())
        self.system_store = self.get_shared_store(VmStore(spath))
        section_getters.append(VmMatcher(self.system_store, name).get_sections)

        for cpath in config_files_in(system_config_dir()):
            store = self.get_shared_store(VmStore(cpath))
            section_getters.append(VmMatcher(store, name).get_sections)

        epath = os.path.join(user_config_dir(),
                             existing_config_file_basename())
        self.existing_store = self.get_shared_store(VmStore(epath))

        super(VmStack, self).__init__(
            section_getters, self.user_store, mutable_section_id=name)

    def iter_sections(self):
        """Iterate all the defined sections.

        We break the lazy loading from byoc here as we want to ensure that
        section referring to a matching host are matched across
        files. Otherwise, a more specific section in a file is masked by a
        shorter section in a previous file.
        """
        # First the cmdline section
        for store, section in self.cmdline_store.get_sections():
            yield store, section

        host_getters = []
        # Now we break the lazy loading: we need to find all matching sections
        # keeping track of the store rank for later sort.
        for rank, section_getter in enumerate(self.sections_def):
            host_getters.extend([(-rank, store, section)
                                 for store, section in section_getter()])
        # We sort on section name then store rank and reverse the whole.
        # Specific section comes first, ties respect store order
        # t[2].id or '' is to cope with None as a section id
        getters = sorted(host_getters, key=lambda t: (t[2].id or '', t[0]),
                         reverse=True)
        for _, store, section in getters:
            yield store, section

        # Finally the existing vm section
        # We want to match only the existing vm, not any other one
        for store, section in stacks.NameMatcher(self.existing_store,
                                                 self.vm_name).get_sections():
            yield store, section

    # FIXME: This should be a DictOption or a NameSpaceOption
    # -- vila 2018-01-08
    def get_nova_creds(self):
        """Get nova credentials from a config.

        This defines the set of options needed to authenticate against nova in
        a single place.

        :raises: byoc.errors.OptionMandatoryValueError if one of the
            options is not set.
        """
        creds = {}
        for k in ('username', 'password', 'tenant_name',
                  'auth_url', 'region_name'):
            opt_name = 'nova.{}'.format(k)
            creds[opt_name] = self.get(opt_name)
        return creds

    # FIXME: This should be a DictOption or a NameSpaceOption
    # -- vila 2021-12-03
    def get_aws_creds(self):
        """Get aws credentials from a config.

        This defines the set of options needed to authenticate against aws in
        a single place.

        :raises: byoc.errors.OptionMandatoryValueError if one of the
            options is not set.
        """
        creds = {}
        for k in ('key', 'secret', 'token', 'region'):
            opt_name = 'aws.{}'.format(k)
            creds[opt_name] = self.get(opt_name)
        return creds

    # FIXME: This should be a DictOption or a NameSpaceOption
    # -- vila 2021-12-03
    def get_scaleway_creds(self):
        """Get scaleway credentials from a config.

        This defines the set of options needed to authenticate against scaleway
        in a single place.

        :raises: byoc.errors.OptionMandatoryValueError if one of the
            options is not set.

        """
        # In theory the token is region agnostic, in practice you can't use
        # compute without a region (the api host name includes the region
        # name).
        creds = {}
        for k in ('access_key', 'token', 'region_name'):
            opt_name = 'scaleway.{}'.format(k)
            creds[opt_name] = self.get(opt_name)
        return creds


class ExistingVmStack(stacks.Stack):
    """Internal stack for defined vms."""

    def __init__(self, name):
        """Make a new stack for an already setup virtual machine.

        :param name: The name of a virtual machine.

        The options are searched only in the ~/.config/byov/existing-vms.conf
        which contains one section per virtual machine.
        """
        dpath = os.path.join(user_config_dir(),
                             existing_config_file_basename())
        store = self.get_shared_store(VmStore(dpath))
        section_getters = [stacks.NameMatcher(store,
                                              name).get_sections]
        super(ExistingVmStack, self).__init__(
            section_getters, store, mutable_section_id=name)


def import_user_byovs():
    # FIXME: This duplicates VmStack definition somehow, byow will need a
    # better way -- vila 2028-10-18
    for d in (os.path.join('.', 'byov.conf.d'),
              os.path.join(user_config_dir(), 'conf.d'),
              os.path.join(system_config_dir(), 'conf.d')):
        byov_path = os.path.abspath(os.path.join(d, 'byov.py'))
        if os.path.exists(byov_path):
            spec = imputil.spec_from_file_location('byov.config.user',
                                                   byov_path)
            module = imputil.module_from_spec(spec)
            spec.loader.exec_module(module)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import os
import unittest

from byov import (
    config,
    errors,
)
from byov.tests import (
    features,
    fixtures,
)
from byov.vms import lxd


class TestStartHook(unittest.TestCase):

    def setUp(self):
        super(TestStartHook, self).setUp()
        fixtures.setup_tests_config(self)
        features.requires_existing_vm(self, 'byov-tests-lxd-ubuntu')
        fixtures.set_uniq_vm_name(self)
        fixtures.override_logging(self)
        # Create a shared config
        conf = config.VmStack(None)
        conf.store._load_from_string('''
[{vm_name}]
vm.name = {vm_name}
vm.class = ephemeral-lxd
vm.backing = byov-tests-lxd-ubuntu
'''.format(vm_name=self.vm_name))
        conf.store.save()
        conf.store.unload()

    def test_empty_hook(self):
        conf = config.VmStack(self.vm_name)
        conf.set('vm.start.hook', '')
        self.vm = lxd.EphemeralLxd(conf)
        self.addCleanup(self.vm.stop)
        # Just a smoke test
        self.vm.start()

    def test_failing_hook(self):
        conf = config.VmStack(self.vm_name)
        conf.set('vm.start.hook', 'i-dont-exist')
        self.vm = lxd.EphemeralLxd(conf)
        with self.assertRaises(errors.CommandError) as cm:
            self.vm.start()
        self.assertEqual(127, cm.exception.retcode)
        self.assertEqual('/bin/sh: 1: i-dont-exist: not found\n',
                         cm.exception.err)

    def test_hook_success(self):
        conf = config.VmStack(self.vm_name)
        conf.set('vm.start.hook', 'touch me')
        self.vm = lxd.EphemeralLxd(conf)
        self.addCleanup(self.vm.stop)
        self.vm.start()
        self.assertTrue(os.path.exists('me'))

    def test_hook_script_success(self):
        conf = config.VmStack(self.vm_name)
        conf.set('content', 'touch me')
        with open('script', 'w') as f:
            f.write('''#!/bin/sh\n{content}\n''')
        os.chmod('script', 0o755)
        conf.set('vm.start.hook', '@./script')
        self.vm = lxd.EphemeralLxd(conf)
        self.addCleanup(self.vm.stop)
        self.vm.start()
        self.assertTrue(os.path.exists('me'))

    def test_hook_script_fails(self):
        conf = config.VmStack(self.vm_name)
        conf.set('content', 'false')
        with open('script', 'w') as f:
            f.write('''#!/bin/sh\n{content}\n''')
        os.chmod('script', 0o755)
        conf.set('vm.start.hook', '@./script')
        self.vm = lxd.EphemeralLxd(conf)
        with self.assertRaises(errors.CommandError) as cm:
            self.vm.start()
        self.assertEqual(1, cm.exception.retcode)


class TestStartHookConfig(unittest.TestCase):

    def setUp(self):
        super(TestStartHookConfig, self).setUp()
        fixtures.setup_tests_config(self)
        features.requires_existing_vm(self, 'byov-tests-lxd-ubuntu')
        fixtures.set_uniq_vm_name(self)
        fixtures.override_logging(self)
        # Create a shared config
        conf = config.VmStack(None)
        conf.store._load_from_string('''
[{vm_name}]
vm.name = {vm_name}
vm.class = ephemeral-lxd
vm.backing = byov-tests-lxd-ubuntu
'''.format(vm_name=self.vm_name))
        conf.store.save()
        conf.store.unload()

    def test_hook_flush_configs(self):
        with open('byov.conf', 'w') as f:
            f.write('''sentinel = now you see me\n''')
        conf = config.VmStack(self.vm_name)
        self.assertEqual('now you see me', conf.get('sentinel'))
        with open('script', 'w') as f:
            f.write('''#!/bin/sh
echo "sentinel = now you dont" > byov.conf
''')
        os.chmod('script', 0o755)
        conf.set('vm.start.hook', '@./script')
        self.vm = lxd.EphemeralLxd(conf)
        self.addCleanup(self.vm.stop)
        self.vm.start()
        self.assertEqual('now you dont', conf.get('sentinel'))



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2016, 2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import sys
import unittest

from byoc import options
from byot import (
    assertions,
    scenarii,
)


from byov import config
from byov.tests import (
    features,
    fixtures,
)

try:
    if sys.version_info < (3,):
        # novaclient doesn't support python3 (yet)
        from byov.vms import nova
except ImportError:
    pass


load_tests = scenarii.load_tests_with_scenarios


@features.requires(features.nova_creds)
class TestByovImageName(unittest.TestCase):

    def test_valid_britney_image(self):
        self.assertEqual(
            'byov/britney/precise-amd64.img',
            nova.byov_image_name('britney', 'precise', 'amd64'))

    def test_valid_cloud_image(self):
        self.assertEqual(
            'byov/cloudimg/precise-amd64.img',
            nova.byov_image_name('cloudimg', 'precise', 'amd64'))

    def test_invalid_image(self):
        with self.assertRaises(ValueError) as cm:
            nova.byov_image_name('I-dont-exist', 'precise', 'amd64')
        self.assertEqual('Invalid image domain', str(cm.exception))


@features.requires(features.nova_creds)
@features.requires(features.nova_compute)
class TestNovaClient(unittest.TestCase):
    """Check the nova client behavior when it encounters exceptions.

    This is achieved by overriding specific methods from NovaClient and
    exercising it through the TestBed methods.
    """

    def setUp(self):
        super(TestNovaClient, self).setUp()
        fixtures.setup_tests_config(self)
        conf = config.VmStack('testing-nova-client')
        conf.set('vm.class', 'nova')
        # Default to precise
        conf.set('vm.release', 'precise')
        self.conf = conf
        fixtures.override_logging(self)

    def get_image_id(self, series, arch):
        return nova.byov_image_name('cloudimg', series, arch)

    def test_retry_is_called(self):
        self.retry_calls = []

        class RetryingNovaClient(nova.NovaClient):

            def __init__(inner, conf, **kwargs):
                # We don't want to wait, it's enough to retry
                super(RetryingNovaClient, inner).__init__(
                    conf, first_wait=0, wait_up_to=0, **kwargs)

            def retry(inner, func, *args, **kwargs):
                self.retry_calls.append((func, args, kwargs))
                return super(RetryingNovaClient, inner).retry(
                    func, *args, **kwargs)

        image_id = self.get_image_id('trusty', 'amd64')
        self.conf.set('nova.image', image_id)
        fixtures.patch(self, nova.NovaServer,
                       'nova_client_class', RetryingNovaClient)
        tb = nova.NovaServer(self.conf)
        self.assertEqual(image_id, tb.find_nova_image().name)
        assertions.assertLength(self, 1, self.retry_calls)

    def test_known_failure_is_retried(self):
        self.nb_calls = 0

        class FailingOnceNovaClient(nova.NovaClient):

            def __init__(inner, conf, **kwargs):
                # We don't want to wait, it's enough to retry
                super(FailingOnceNovaClient, inner).__init__(
                    conf, first_wait=0, wait_up_to=0, retries=1,
                    **kwargs)

            def fail_once(inner):
                self.nb_calls += 1
                if self.nb_calls == 1:
                    raise nova.client.requests.ConnectionError()
                else:
                    return inner.nova.flavors.list()

            def flavors_list(inner):
                return inner.retry(inner.fail_once)

        fixtures.patch(self, nova.NovaServer,
                       'nova_client_class', FailingOnceNovaClient)
        tb = nova.NovaServer(self.conf)
        tb.find_flavor()
        self.assertEqual(2, self.nb_calls)

    def test_unknown_failure_is_raised(self):

        class FailingNovaClient(nova.NovaClient):

            def __init__(inner, conf, **kwargs):
                # We don't want to wait, it's enough to retry
                super(FailingNovaClient, inner).__init__(
                    conf, first_wait=0, wait_up_to=0,
                    **kwargs)

            def fail(inner):
                raise AssertionError('Boom!')

            def flavors_list(inner):
                return inner.retry(inner.fail)

        fixtures.patch(self, nova.NovaServer,
                       'nova_client_class', FailingNovaClient)
        tb = nova.NovaServer(self.conf)
        # This mimics what will happen when we encounter unknown transient
        # failures we want to catch: an exception will bubble up and we'll have
        # to add it to NovaClient.retry().
        with self.assertRaises(nova.NovaServerException) as cm:
            tb.find_flavor()
        self.assertEqual('fail failed', str(cm.exception))


@features.requires(features.nova_creds)
@features.requires(features.nova_compute)
class TestTestbed(unittest.TestCase):

    def setUp(self):
        super(TestTestbed, self).setUp()
        fixtures.setup_tests_config(self)
        conf = config.VmStack('testing-testbed')
        conf.set('vm.class', 'nova')
        conf.set('vm.name', 'testing-testbed')
        # Default to precise
        conf.set('vm.release', 'precise')
        self.conf = conf
        fixtures.override_logging(self)

    def get_image_id(self, series='precise', arch='amd64'):
        return nova.byov_image_name('cloudimg', series, arch)

    def test_create_no_image(self):
        tb = nova.NovaServer(self.conf)
        with self.assertRaises(options.errors.OptionMandatoryValueError) as cm:
            tb.setup()
        self.assertEqual('nova.image must be set.', str(cm.exception))

    def test_create_no_flavor(self):
        tb = nova.NovaServer(self.conf)
        tb.conf.set('nova.flavors', '')
        with self.assertRaises(nova.NovaServerException) as cm:
            tb.setup()
        self.assertEqual('nova.flavors must be set', str(cm.exception))

    def test_create_unknown_image(self):
        image_name = "I don't exist and eat kittens"
        self.conf.set('nova.image', image_name)
        tb = nova.NovaServer(self.conf)
        with self.assertRaises(nova.NovaServerException) as cm:
            tb.setup()
        self.assertEqual('Image "{}" cannot be found'.format(image_name),
                         str(cm.exception))

    def test_create_unknown_flavor(self):
        flavors = "I don't exist and eat kittens"
        self.conf.set('nova.flavors', flavors)
        tb = nova.NovaServer(self.conf)
        with self.assertRaises(nova.NovaServerException) as cm:
            tb.setup()
        self.assertEqual('None of [{}] can be found'.format(flavors),
                         str(cm.exception))

    def test_wait_for_instance_fails(self):
        self.conf.set('nova.image', self.get_image_id())
        # Force a 0 timeout so the instance can't finish booting
        self.conf.set('nova.boot_timeout', '0')
        tb = nova.NovaServer(self.conf)
        self.addCleanup(tb.teardown)
        with self.assertRaises(nova.NovaServerException) as cm:
            tb.setup()
        msg = 'Instance {} never came up (last status: BUILD)'
        msg = msg.format(tb.instance.id)
        self.assertEqual(msg, str(cm.exception))

    def test_wait_for_instance_errors(self):
        self.conf.set('nova.image', self.get_image_id())
        tb = nova.NovaServer(self.conf)
        self.addCleanup(tb.teardown)

        def update_instance_to_error():
            # Fake an instance starting in error state
            tb.instance.status = 'ERROR'
            return True
        tb.update_instance = update_instance_to_error
        with self.assertRaises(nova.NovaServerException) as cm:
            tb.setup()
        msg = 'Instance {} never came up (last status: ERROR)'
        msg = msg.format(tb.instance.id)
        self.assertEqual(msg, str(cm.exception))


# FIXME: parametrized by distribution ? -- vila 2017-12-06
@features.requires(features.nova_creds)
@features.requires(features.nova_compute)
class TestUsableTestbed(unittest.TestCase):

    scenarios = scenarii.multiply_scenarios(
        # series
        ([('precise', dict(series='precise', result='skip')),
          ('trusty', dict(series='trusty', result='pass')),
          ('xenial', dict(series='xenial', result='pass')),
          ('yakkety', dict(series='yakkety', result='pass')),
          ('zesty', dict(series='zesty', result='pass'))]),
        # architectures
        ([('amd64', dict(arch='amd64')), ('i386', dict(arch='i386'))]))

    def setUp(self):
        super(TestUsableTestbed, self).setUp()
        fixtures.setup_tests_config(self)
        tb_name = 'testing-testbed-{}-{}'.format(self.series, self.arch)
        conf = config.VmStack(tb_name)
        conf.set('vm.class', 'nova')
        conf.set('vm.name', tb_name)
        conf.set('vm.release', self.series)
        self.conf = conf
        fixtures.override_logging(self)

    def get_image_id(self):
        return nova.byov_image_name('cloudimg', self.series, self.arch)

    def test_create_usable_testbed(self):
        self.conf.set('nova.image', self.get_image_id())
        tb = nova.NovaServer(self.conf)
        self.addCleanup(tb.teardown)
        tb.setup()
        # We should be able to ssh with the right user
        retcode, out, err = tb.shell_captured('whoami')
        self.assertEqual(0, retcode)
        self.assertEqual(self.conf.get('vm.user') + '\n', out)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2015, 2016, 2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import hashlib
import logging
import os

from byov import (
    config,
    errors,
    subprocesses,
    vms,
)


logger = logging.getLogger(__name__)


def next_subids(typ, etc_dir=None):
    if etc_dir is None:
        etc_dir = '/etc'
    fname = 'sub{}'.format(typ)
    with open(os.path.join(etc_dir, fname)) as f:
        for line in f.readlines():
            if line.startswith('#'):
                continue
            try:
                user, first, nb = line.strip().split(':')
                first = int(first)
                nb = int(nb)
            except ValueError:
                continue
            yield user, first, nb


def check_id_for(name, the_id, typ, etc_dir=None):
    for user, first, nb in next_subids(typ, etc_dir):
        if user == name and first <= the_id < (first + nb):
            return True
    return False


def available_ids(name, typ, etc_dir=None):
    total = 0
    for user, _, nb in next_subids(typ, etc_dir):
        if user == name:
            # FIXME: Recovering range definitions ? -- vila 2017-01-19
            total += nb
    return total


def user_can_nest(user, level, etc_dir=None):
    # For each level of nesting, a full additional range is needed
    # No nesting still requires a full range
    needed_ids = (level + 1) * 65536
    uids = available_ids(user, 'uid', etc_dir)
    gids = available_ids(user, 'gid', etc_dir)
    return (uids >= needed_ids and gids >= needed_ids)


def check_nesting(level, etc_dir=None):
    root = user_can_nest('root', level, etc_dir)
    lxd = user_can_nest('lxd', level, etc_dir)
    return root and lxd


def lxd_info(vm_name, source=None):
    info = dict(state='UNKNOWN')
    if source is None:
        info_command = ['lxc', 'info', vm_name]
        try:
            _, source, err = subprocesses.run(info_command)
        except errors.CommandError as e:
            if e.err.lower() == 'error: not found\n':
                return info
            else:
                raise
    scanning_ips = False
    for l in source.splitlines():
        if l.startswith('Resources:'):
            # Nothing for us from there
            break
        elif scanning_ips:
            if l == '(none)':
                # interfaces are not setup yet
                continue
            details = l.split('\t')
            try:
                ifce, proto, ip = details[0:3]
                ifce = ifce[0:-1].strip()
            except ValueError:
                raise ValueError('Unexpected IP line: {}'.format(l))
            # FIXME: Assume eth0 until we can do better -- vila 2015-12-22
            if 'eth0' in ifce and proto == 'inet':
                info['ip'] = ip
        elif l.startswith('Status:'):
            _, s = l.split()
            info['state'] = s.upper()
        elif l.startswith('Pid:'):
            _, pid = l.split()
            info['pid'] = pid
        elif l.startswith('Ips:'):
            scanning_ips = True
    return info


class Lxd(vms.VM):

    setup_ip_timeouts = 'lxd.setup_ip.timeouts'
    setup_ssh_timeouts = 'lxd.setup_ssh.timeouts'
    existing_vm_options = vms.VM.existing_vm_options + ['lxd.remote']

    def get_lxd_name(self):
        name = self.conf.get('vm.name')
        remote = self.conf.get('lxd.remote')
        if remote:
            name = '{}:{}'.format(remote, name)
        return name

    def state(self):
        # man lxc(7) defines the possible states as: STOPPED, STARTING,
        # RUNNING, ABORTING, STOPPING. We add UNKNOWN.
        info = lxd_info(self.get_lxd_name())
        return info['state']

    def init(self):
        init_command = ['lxc', 'init', self.conf.get('lxd.image'),
                        self.get_lxd_name()]
        profiles = self.conf.get('lxd.profiles')
        if profiles:
            for p in profiles:
                init_command.extend(['-p', p])
        nesting = self.conf.get('lxd.nesting')
        if not check_nesting(nesting):
            raise errors.ByovError(
                'Lxd needs more ids for {nesting} levels of nesting',
                nesting=nesting)
        if nesting:
            init_command.extend(
                ['--config', 'security.nesting=True'])
        if self.conf.get('lxd.user_mounts'):
            self.check_subids_for_mounts(os.getuid(), os.getgid())
        mac_address = self.conf.get('lxd.mac.address')
        if mac_address:
            # FIXME: Can we do better than eth0 ? -- vila 2018-03-09
            init_command.extend(
                ['--config', 'volatile.eth0.hwaddr={}'.format(mac_address)])
        if self.conf.get('lxd.config.boot.autostart'):
            init_command.extend(['--config', 'boot.autostart=true'])

        logger.debug('Initializing {} lxd container with {}'.format(
            self.conf.get('vm.name'), ' '.join(init_command)))
        # FIXME: Log out & err ? -- vila 2016-01-05
        # FIXME: This can hang IRL (apparently when a new image needs to be
        # downloaded requiring an lxd restart, but this cannot be reliably
        # reproduced so far) and should be guarded by a timeout
        # -- vila 2016-11-16
        return subprocesses.run(init_command)

    def set_cloud_init_config(self):
        self.create_user_data()
        config_command = ['lxc', 'config', 'set', self.get_lxd_name(),
                          'user.user-data', '-']
        logger.debug('Configuring {} lxd container with {}'.format(
            self.conf.get('vm.name'), ' '.join(config_command)))
        subprocesses.run(config_command, cmd_input=self.ci_user_data.dump())

    def check_subids_for_mounts(self, uid, gid, etc_dir=None):
        if not check_id_for('root', uid, 'uid', etc_dir):
            raise errors.SubIdError('uid', uid)
        if not check_id_for('root', gid, 'gid', etc_dir):
            raise errors.SubIdError('gid', gid)

    def get_mount_id_map(self):
        # Declare the needed id mapping.
        huid = self.conf.get('lxd.user_mounts.host.uid')
        hgid = self.conf.get('lxd.user_mounts.host.gid')
        vuid = self.conf.get('lxd.user_mounts.container.uid')
        vgid = self.conf.get('lxd.user_mounts.container.gid')
        # We can't use 'both {host uid/gid} {vm uid/gid}' as uid and gid
        # may differ.
        mapping_template = '{map} {host_id}-{host_id} {vm_id}-{vm_id}'
        uid_map = mapping_template.format(map='uid', host_id=huid, vm_id=vuid)
        gid_map = mapping_template.format(map='gid', host_id=hgid, vm_id=vgid)
        return uid_map, gid_map

    def set_user_mounts(self):
        mounts = self.conf.get('lxd.user_mounts')
        if not mounts:
            return

        config_command = ['lxc', 'config', 'set', self.get_lxd_name(),
                          'raw.idmap', '-']
        mapping = ''.join([m + '\n' for m in self.get_mount_id_map()])
        logger.debug('Configuring {} lxd container with {}'.format(
            self.conf.get('vm.name'),
            ' '.join(config_command) + '\n' + mapping))
        subprocesses.run(config_command, cmd_input=mapping)

        base_command = ['lxc', 'config', 'device', 'add',
                        self.get_lxd_name()]
        for rank, (host_path, vm_path) in enumerate(mounts):
            disk_def = ['byovdisk{}'.format(rank),
                        'disk', 'source={}'.format(host_path),
                        'path={}'.format(vm_path)]
            config_command = base_command + disk_def
            logger.debug('Adding mount for {} with {}'.format(
                self.conf.get('vm.name'), ' '.join(config_command)))
            subprocesses.run(config_command)

    def setup(self):
        logger.info('Setting up {} lxd container...'.format(
            self.conf.get('vm.name')))
        self.init()
        self.set_cloud_init_config()
        self.set_user_mounts()
        self.start()  # calls wait_for_ssh()
        self.wait_for_cloud_init()
        self.setup_over_ssh()

    def hash_setup(self):
        hasher = hashlib.md5()
        # Seed with the digest from the base class to capture common options
        hasher.update(super(Lxd, self).hash_setup().encode('utf8'))
        options = self.conf.get('lxd.setup.digest.options')
        if options:
            for opt in options:
                value = self.conf.get(opt, convert=False)
                self.hash_value(hasher, value)
        return hasher.hexdigest()

    def discover_ip(self):
        info = lxd_info(self.get_lxd_name())
        if info['state'] != 'RUNNING' or 'ip' not in info:
            raise errors.ByovError(
                'Lxd {name} has not provided an IP yet: {info}',
                name=self.get_lxd_name(), info=info)
        return info['ip']

    def wait_for_cloud_init(self):
        self._wait_for_cloud_init('lxd.cloud_init.timeouts')

    def start(self):
        super(Lxd, self).start()
        logger.info('Starting {} lxd container...'.format(
            self.conf.get('vm.name')))
        start_command = ['lxc', 'start', self.get_lxd_name()]
        subprocesses.run(start_command)
        self.wait_for_ip()
        self.wait_for_ssh()
        self.save_existing_config()

    def stop(self):
        logger.info('Stopping {} lxd container...'.format(
            self.conf.get('vm.name')))
        # FIXME: Seen failing with:
        # command: lxc stop --force TestSetupHook-test-hook-success-12702
        # retcode: 1
        # stdout:
        # stderr: Error: websocket: close 1006 (abnormal closure): \
        #                unexpected EOF
        # -- vila 2019-10-16
        stop_command = ['lxc', 'stop', '--force', self.get_lxd_name()]
        subprocesses.run(stop_command)

    def publish(self):
        img_name = self.conf.get('vm.published_as')
        logger.info('Publishing {} lxd image from {}...'.format(
            img_name, self.conf.get('vm.name')))
        publish_command = ['lxc', 'publish', self.get_lxd_name(),
                           '--alias', img_name]
        # FIXME: Seen failing with:
        #   command: lxc publish TestPublish-test-publish-7621 \
        #                --alias TestPublish-test-publish-7621
        # retcode: 1
        # output:
        # error: error: websocket: close 1006 (abnormal closure): \
        #               unexpected EOF
        # -- vila 2018-01-17
        subprocesses.run(publish_command)

    def unpublish(self):
        img_name = self.conf.get('vm.published_as')
        logger.info('Un-publishing {} lxd image from {}...'.format(
            img_name, self.conf.get('vm.name')))
        unpublish_command = ['lxc', 'image', 'delete', img_name]
        subprocesses.run(unpublish_command)

    def teardown(self, force=False):
        logger.info('Tearing down {} lxd container...'.format(
            self.conf.get('vm.name')))
        if force and self.state() == 'RUNNING':
            self.stop()
        teardown_command = ['lxc', 'delete', '--force',
                            self.get_lxd_name()]
        subprocesses.run(teardown_command)
        super(Lxd, self).teardown()


class EphemeralLxd(Lxd):

    def setup(self):
        raise NotImplementedError(self.setup)

    def teardown(self):
        raise NotImplementedError(self.teardown)

    def start(self):
        # /!\ We don't call super(EphemeralLxd,self) ! We just want the Vm one.
        super(Lxd, self).start()
        backing_conf = config.VmStack(self.conf.get('vm.backing'))
        # Inherit basic values from the backing vm. Those /could/ be changed by
        # the user but won't make sense.
        self.conf.set('vm.distribution', backing_conf.get('vm.distribution'))
        self.conf.set('vm.release', backing_conf.get('vm.release'))
        self.conf.set('vm.architecture', backing_conf.get('vm.architecture'))
        # Back to ephemeral specifics:
        mounts = self.conf.get('lxd.user_mounts')
        if mounts and backing_conf.get('lxd.user_mounts'):
            # FIXME: We could have mounts in both backing and the ephemeral but
            # that requires checking which mounts are already set to not add
            # the same volume twice (and also find a better naming scheme to
            # not conflict (byovdiskN needs at least to start above the
            # existing ones) -- vila 2017-01-12

            # FIXME: The solution is one vm.backing.conf.get('lxd.user_mounts')
            # away -- vila 2017-02-04
            raise errors.ByovError(
                'Backing vm {backing} already has mounts',
                backing=self.conf.get('vm.backing'))
        logger.info('Starting {} ephemeral lxd container...'.format(
            self.conf.get('vm.name')))
        # FIXME: This should get vm.published_as from the backing vm
        # -- vila 2019-11-17
        copy_command = ['lxc', 'copy', self.conf.get('vm.backing'),
                        self.get_lxd_name(),
                        '--ephemeral']
        subprocesses.run(copy_command)
        self.set_user_mounts()
        start_command = ['lxc', 'start', self.get_lxd_name()]
        subprocesses.run(start_command)
        self.wait_for_ip()
        self.wait_for_ssh()
        self.setup_over_ssh()

    def stop(self):
        super(EphemeralLxd, self).stop()
        # Stopping an ephemeral deletes it
        self.cleanup_configs()

# MISSINGTEST: Experiment and run tests in qemu where lxd is not installed

# /etc/cloud/cloud.cfg is the defaut config for cloud-init

# /etc/cloud/cloud.cfg.d/90_dpkg.cfg may be worth overriding to avoid searching
# for clouds we don't use



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2022 Vincent Ladeuil.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import io
import unittest

from byoc import errors
from byov import (
    config,
)
from byov.tests import (
    features,
    fixtures,
)
from byov.vms import docker


@features.requires(features.docker_client_feature)
class TestContainerInspect(unittest.TestCase):

    def setUp(self):
        super().setUp()
        fixtures.setup_tests_config(self)
        fixtures.set_uniq_vm_name(self)
        # Create a shared config
        conf = config.VmStack(self.vm_name)
        conf.set('vm.name', self.vm_name)
        conf.set('vm.class', 'docker')
        conf.set('docker.image', 'alpine')
        conf.store.save()
        conf.store.unload()
        fixtures.override_logging(self)

    def define_docker(self, definition):
        with io.open('./Dockerfile', 'w', encoding='utf8') as f:
            f.write(definition)
        vm = docker.Docker(config.VmStack(self.vm_name))
        return vm

    def test_unknown(self):
        info = docker.container_info('idontexist')
        self.assertEqual('UNKNOWN', info['state'])

    def test_running(self):
        vm = self.define_docker('''
FROM alpine
# Force the container to stay up once started
ENTRYPOINT ["tail", "-f", "/dev/null"]
''')
        self.addCleanup(vm.teardown)
        vm.setup()
        self.addCleanup(vm.stop)
        vm.start()
        self.assertEqual('RUNNING', vm.state())

    def test_stopped(self):
        vm = self.define_docker('''
FROM alpine
''')
        self.addCleanup(vm.teardown)
        vm.setup()
        vm.stop()
        self.assertEqual('STOPPED', vm.state())


# FIXME: parametrized by distribution ? -- vila 2017-12-06
@features.requires(features.docker_client_feature)
class TestSetup(unittest.TestCase):

    def setUp(self):
        super(TestSetup, self).setUp()
        fixtures.setup_tests_config(self)
        fixtures.set_uniq_vm_name(self)
        # Create a shared config
        conf = config.VmStack(self.vm_name)
        conf.set('vm.name', self.vm_name)
        conf.set('vm.class', 'docker')
        conf.set('vm.release', 'focal')
        conf.set('vm.architecture', 'amd64')
        conf.store.save()
        conf.store.unload()
        with io.open('./Dockerfile', 'w', encoding='utf8') as f:
            f.write('''
FROM alpine
# Force the container to stay up once started
ENTRYPOINT ["tail", "-f", "/dev/null"]
''')

    def test_usable(self):
        vm = docker.Docker(config.VmStack(self.vm_name))
        self.addCleanup(vm.teardown)
        self.addCleanup(vm.stop)
        vm.setup()
        # We should be able to ssh with the right user
        ret, out, err = vm.shell_captured('whoami')
        self.assertEqual(0, ret)
        # Defaulting to root is bad, even if it's docker's fault
        self.assertEqual('root' + '\r\n', out)


class TestDockerMountsOption(unittest.TestCase):

    maxDiff = None

    def setUp(self):
        super().setUp()
        fixtures.setup_tests_config(self)
        fixtures.set_uniq_vm_name(self)
        # Create a shared config
        conf = config.VmStack(self.vm_name)
        conf.set('vm.name', self.vm_name)
        conf.set('vm.class', 'docker')
        conf.set('vm.release', 'focal')
        conf.set('vm.architecture', 'amd64')
        conf.store.save()
        conf.store.unload()
        self.conf = conf

    def _test_invalid(self, val):
        self.conf.set('docker.mounts', val)
        with self.assertRaises(errors.OptionValueError) as cm:
            self.conf.get('docker.mounts')
        self.assertEqual(val, cm.exception.value)

    def test_invalid_type(self):
        self._test_invalid('wrong:~:/there')

    def test_invalid_readonly(self):
        self._test_invalid('bind:~:/home:read')

    def test_invalid_share(self):
        self._test_invalid('bind:~:/home:readonly:notashare')

    def test_bind_mount(self):
        self.conf.set('docker.mounts', 'bind:~:/home:readonly:rslave')
        self.assertEqual(
            ['--mount=type=bind,src={},dst=/home'
             ',readonly,bind-propagation=rslave'.format(
                 self.home_dir)],
            self.conf.get('docker.mounts'))

    def test_tmpfs_mount(self):
        self.conf.set('docker.mounts', 'tmpfs:/tmp:65536:1777')
        self.assertEqual(
            ['--mount=type=tmpfs,dst=/tmp,tmpfs-size=65536,tmpfs-mode=1777'],
            self.conf.get('docker.mounts'))

    def test_volume_mount(self):
        self.conf.set('docker.mounts', 'volume:volname:/there')
        self.assertEqual(
            ['--mount=type=volume,src=volname,dst=/there'],
            self.conf.get('docker.mounts'))



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2021 Vincent Ladeuil.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from byot import (
    scenarii,
)


from byov import (
    config,
    options,
)
from byov.tests import (
    features,
    fixtures,
)

try:
    from byov.vms import ec2
except ImportError:
    pass


load_tests = scenarii.load_tests_with_scenarios


@features.requires(features.ec2_creds)
@features.requires(features.ec2_boto3)
class TestTags(unittest.TestCase):
    """Test all tags used on ec2 resources to monitor usage and leaks."""
    maxDiff = None

    def setUp(self):
        super().setUp()
        fixtures.setup_tests_config(self)
        fixtures.set_uniq_vm_name(self)
        conf = config.VmStack(self.vm_name)
        conf.set('vm.class', 'ec2')
        conf.set('vm.name', self.vm_name)
        conf.set('vm.distribution', 'ubuntu')
        conf.set('vm.release', 'focal')
        self.conf = conf
        fixtures.override_logging(self)
        self.vm = ec2.Ec2Server(self.conf)

    def test_instance_tags(self):
        tags = self.vm.instance_tags()
        self.assertEqual([{'ResourceType': 'instance',
                           'Tags':
                           [{'Key': 'Name', 'Value': self.vm_name},
                            {'Key': 'test.id', 'Value': self.id()}]}],
                         tags)

    def test_image_tags(self):
        tags = self.vm.image_tags()
        self.assertEqual([{'ResourceType': 'image',
                           'Tags':
                           [{'Key': 'Name', 'Value': self.vm_name},
                            {'Key': 'test.id', 'Value': self.id()}]},
                          {'ResourceType': 'snapshot',
                           'Tags':
                           [{'Key': 'Name', 'Value': self.vm_name},
                            {'Key': 'test.id', 'Value': self.id()}]}],
                         tags)


@features.requires(features.ec2_creds)
@features.requires(features.ec2_boto3)
class TestDescribeInstanceTypes(unittest.TestCase):

    def setUp(self):
        super().setUp()
        fixtures.setup_tests_config(self)
        conf = config.VmStack('testing-testbed')
        conf.set('vm.class', 'ec2')
        conf.set('vm.name', 'testing-testbed')
        conf.set('vm.distribution', 'ubuntu')
        conf.set('vm.release', 'focal')
        self.conf = conf
        fixtures.override_logging(self)
        self.ec2_client = ec2.Ec2Client(self.conf)

    def test_unknown_type(self):
        with self.assertRaises(ec2.Ec2ServerException) as cm:
            self.ec2_client.instance_types(['unknown'], free_tier=True)
        boto_exc = cm.exception.boto_exc
        self.assertTrue(boto_exc.args[0].endswith('do not exist: [unknown]'))
        self.assertEqual('DescribeInstanceTypes', boto_exc.operation_name)
        resp = boto_exc.response
        self.assertEqual(400, resp['ResponseMetadata']['HTTPStatusCode'])

    def test_free(self):
        # Careful with an empty list below we get all existing types (and then
        # filter the free ones).
        types = self.ec2_client.instance_types([], free_tier=True)
        self.assertEqual(['t1.micro', 't2.micro'], types)

    def test_all(self):
        # Careful with an empty list below we get all existing types (and then
        # filter out the free ones).
        types = self.ec2_client.instance_types([], free_tier=False)
        assert 400 < len(types)
        assert 't1.micro' not in types
        assert 't2.micro' not in types


@features.requires(features.ec2_creds)
@features.requires(features.ec2_boto3)
class TestDescribeImages(unittest.TestCase):

    def setUp(self):
        super().setUp()
        fixtures.setup_tests_config(self)
        conf = config.VmStack('testing-testbed')
        conf.set('vm.class', 'ec2')
        conf.set('vm.name', 'testing-testbed')
        conf.set('vm.distribution', 'ubuntu')
        conf.set('vm.release', 'focal')
        conf.set('vm.architecture', 'amd64')
        self.conf = conf
        fixtures.override_logging(self)
        self.ec2_client = ec2.Ec2Client(self.conf)

    def images(self, image_id=None):
        arch = ec2.ec2_architectures[self.conf.get('vm.architecture')]
        owners = ec2.ec2_image_owners.get(self.conf.get('vm.distribution'),
                                          ['self'])
        return self.ec2_client.images(arch, owners, image_id)

    def test_malformed_image(self):
        with self.assertRaises(ec2.Ec2ServerException) as cm:
            self.images(image_id='not-an-ami')
        boto_exc = cm.exception.boto_exc
        self.assertTrue('InvalidAMIID.Malformed' in boto_exc.args[0])
        self.assertTrue('Invalid id: "not-an-ami"' in boto_exc.args[0])
        self.assertEqual('DescribeImages', boto_exc.operation_name)
        resp = boto_exc.response
        self.assertEqual(400, resp['ResponseMetadata']['HTTPStatusCode'])

    def test_unknown_image(self):
        with self.assertRaises(ec2.Ec2ServerException) as cm:
            # Empirically found a well formed but not existing image for this
            # test
            self.images(image_id='ami-01fffffffffffffff')
        boto_exc = cm.exception.boto_exc
        self.assertTrue('InvalidAMIID.NotFound' in boto_exc.args[0])
        self.assertTrue(boto_exc.args[0].endswith(
            "'[ami-01fffffffffffffff]' does not exist"))
        self.assertEqual('DescribeImages', boto_exc.operation_name)
        resp = boto_exc.response
        # 404 sounds more appropriate but ec2 says 400
        self.assertEqual(400, resp['ResponseMetadata']['HTTPStatusCode'])

    def test_some_images(self):
        # At the time of writing this test, there are 156.100 available images
        # 116.000 still for virt. type hvm and x86_64 arch
        # 116.00 still for type machine
        # 18.000 when filtering on owner (self, amazon, canonical, debian and
        #        redhat)
        #  6.800 when filtering on owner (self, canonical)
        images = self.images()
        self.assertTrue(6800 < len(images))

    def test_known_image_id(self):
        # was ubuntu/images/hvm-ssd/ubuntu-focal-20.04-amd64-server-20211129
        # on 2021-11-30
        # FIXME: We need either a stable id or be prepared to update the test
        # when the used one become unavailable -- vila 2021-11-30
        images = self.images(image_id='ami-04505e74c0741db8d')
        self.assertEqual(1, len(images))


@features.requires(features.ec2_creds)
@features.requires(features.ec2_boto3)
class TestFindEc2Image(unittest.TestCase):

    def setUp(self):
        super().setUp()
        fixtures.setup_tests_config(self)
        conf = config.VmStack('testing-testbed')
        conf.set('vm.class', 'ec2')
        conf.set('vm.name', 'testing-testbed')
        conf.set('vm.distribution', 'ubuntu')
        conf.set('vm.architecture', 'amd64')
        conf.set('vm.release', 'focal')
        self.conf = conf
        fixtures.override_logging(self)
        self.tb = ec2.Ec2Server(self.conf)

    def test_no_image(self):
        with self.assertRaises(options.OptionMandatoryValueError) as cm:
            self.tb.find_ec2_image()
        self.assertEqual('ec2.image must be set.', str(cm.exception))

    def test_known_image(self):
        # FIXME: Use a better value for ec2.image -- vila 2021-12-08
        self.conf.set('ec2.image', '^ubuntu')
        expected = 'ami-04505e74c0741db8d'
        self.conf.set('ec2.image.id', expected)
        image_id = self.tb.find_ec2_image()['ImageId']
        self.assertEqual(expected, image_id)


@features.requires(features.ec2_creds)
@features.requires(features.ec2_boto3)
class TestFindEc2ImageByDistribution(unittest.TestCase):

    scenarios = scenarii.multiply_scenarios(
        # distro/series
        ([('ubuntu,focal', dict(distro='ubuntu', release='focal')),
          ('ubuntu,impish', dict(distro='ubuntu', release='impish')),
          ('debian,10', dict(distro='debian', release='10')),
          ('debian,11', dict(distro='debian', release='11')),
          ('al2', dict(distro='amazon', release='2'))]),
        # architectures
        ([('amd64', dict(arch='amd64'))]))

    def setUp(self):
        super().setUp()
        fixtures.setup_tests_config(self)
        conf = config.VmStack('testing-testbed')
        conf.set('vm.class', 'ec2')
        conf.set('vm.name', 'testing-testbed')
        conf.set('vm.distribution', self.distro)
        conf.set('vm.architecture', self.arch)
        conf.set('vm.release', self.release)
        conf.set('ec2.image', options.ec2_image_re_registry.get(self.distro))
        conf.set('aws.region', 'us-east-1')
        self.conf = conf
        fixtures.override_logging(self)
        self.tb = ec2.Ec2Server(self.conf)

    def test_default_image(self):
        image = self.tb.find_ec2_image()
        self.assertTrue(image['ImageId'] is not None)
        # FIXME: stricter checks around distribution/series/arch ?
        # -- vila 2022-02-11


@features.requires(features.ec2_creds)
@features.requires(features.ec2_boto3)
class TestTestbed(unittest.TestCase):

    def setUp(self):
        super(TestTestbed, self).setUp()
        fixtures.setup_tests_config(self)
        conf = config.VmStack('testing-testbed')
        conf.set('vm.class', 'ec2')
        conf.set('vm.name', 'testing-testbed')
        conf.set('vm.distribution', 'ubuntu')
        conf.set('vm.release', 'focal')
        conf.set('vm.architecture', 'amd64')
        self.conf = conf
        fixtures.override_logging(self)

    # FIXME: This should fail. It's valid to not set ec2.image and relies on
    # ec2.distribution.images instead -- vila 2022-02-11
    def test_no_ec2_image(self):
        tb = ec2.Ec2Server(self.conf)
        self.addCleanup(tb.teardown)
        with self.assertRaises(options.OptionMandatoryValueError) as cm:
            tb.setup()
        self.assertEqual('ec2.image must be set.', str(cm.exception))

    def test_no_instance_types(self):
        tb = ec2.Ec2Server(self.conf)
        self.addCleanup(tb.teardown)
        tb.conf.set('ec2.instance.types', '')
        with self.assertRaises(ec2.Ec2ServerException) as cm:
            tb.setup()
        self.assertEqual('ec2.instance.types must be set', str(cm.exception))

    def test_unknown_image(self):
        self.conf.set('ec2.image', '')
        image_name = 'ami-01fffffffffffffff'
        self.conf.set('ec2.image.id', image_name)
        tb = ec2.Ec2Server(self.conf)
        self.addCleanup(tb.teardown)
        with self.assertRaises(ec2.Ec2ServerException) as cm:
            tb.setup()
        msg_end = "The image id '[{}]' does not exist".format(image_name)
        self.assertTrue(str(cm.exception).endswith(msg_end), cm.exception)

    def test_unknown_type(self):
        itypes = 'I-dont-exist'
        self.conf.set('ec2.instance.types', itypes)
        tb = ec2.Ec2Server(self.conf)
        self.addCleanup(tb.teardown)
        with self.assertRaises(ec2.Ec2ServerException) as cm:
            tb.setup()
        msg_end = "instance types do not exist: [{}]".format(itypes)
        self.assertTrue(str(cm.exception).endswith(msg_end), cm.exception)

    def test_wait_for_instance_fails(self):
        self.conf.set('ec2.image', '')
        # Force a 0 timeout so the instance can't finish booting
        self.conf.set('ec2.boot.timeouts', '0, 0, 0')
        tb = ec2.Ec2Server(self.conf)
        self.addCleanup(tb.teardown)
        with self.assertRaises(ec2.Ec2ServerException) as cm:
            tb.setup()
        msg = ("{} never reached ['RUNNING']"
               " (was: UNKNOWN last)")
        msg = msg.format(tb.conf.get('ec2.instance.id'))
        self.assertEqual(msg, str(cm.exception))


# FIXME: parametrized by distribution ? -- vila 2017-12-06
@features.requires(features.ec2_creds)
@features.requires(features.ec2_boto3)
class TestUsableTestbed(unittest.TestCase):

    scenarios = scenarii.multiply_scenarios(
        # series
        ([('focal', dict(release='focal'))]),
        # architectures
        ([('amd64', dict(arch='amd64'))]))

    def setUp(self):
        super(TestUsableTestbed, self).setUp()
        fixtures.setup_tests_config(self)
        tb_name = 'testing-testbed-{}-{}'.format(self.release, self.arch)
        conf = config.VmStack(tb_name)
        conf.set('vm.class', 'ec2')
        conf.set('vm.name', tb_name)
        conf.set('vm.release', self.release)
        conf.set('vm.architecture', self.arch)
        conf.set('ec2.image', '')
        self.conf = conf
        fixtures.override_logging(self)

    def test_create_usable_testbed(self):
        tb = ec2.Ec2Server(self.conf)
        self.addCleanup(tb.teardown)
        tb.setup()
        # We should be able to ssh with the right user
        retcode, out, err = tb.shell_captured('whoami')
        self.assertEqual(0, retcode)
        self.assertEqual(self.conf.get('vm.user') + '\n', out)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2015, 2016, 2017 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import unittest

from byot import assertions

from byov import timeouts


class TestExponentialBackoff(unittest.TestCase):

    def test_no_timeouts(self):
        timo = timeouts.ExponentialBackoff(1, 2, 0)
        assertions.assertLength(self, 0, list(timo))

    def test_single_retry(self):
        eb = timeouts.ExponentialBackoff(0, 2, 1)
        timo = list(eb)
        assertions.assertLength(self, 2, timo)

    def test_two_retries(self):
        eb = timeouts.ExponentialBackoff(0, 2, 2)
        timo = list(eb)
        assertions.assertLength(self, 3, timo)

    def test_up_to_5mins(self):
        timo = timeouts.ExponentialBackoff(12, 300, 10)
        values = list(timo)
        # There is one more timeout than retries
        assertions.assertLength(self, 11, values)
        # The first duration is explicit
        self.assertEqual(12, values[0])
        # The sum of wait times is equal to up_to
        self.assertEqual(300, sum(values))



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2019 Vincent Ladeuil.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import os
import unittest


from byov import (
    config,
    errors,
)
from byov.tests import (
    features,
    fixtures,
)
from byov.vms import qemu


# FIXME: Not used ? -- vila 2019-11-15
def requires_known_reference_image(test):
    """Skip test if a known reference image is not provided locally.

    :note: This should be called early during setUp so the user configuration
        is still available (i.e. the test has not yet been isolated from disk).

    :param test: The TestCase requiring the image.
    """
    # We need a pre-seeded download cache from the user running the tests
    # as downloading the cloud image is too long.
    user_conf = config.VmStack(None)
    download_cache = user_conf.get('qemu.download.dir')
    # FIXME: Testing directory existence will offer an opt-in path for limiting
    # tests -- vila 2019-11-12
    if download_cache is None:
        test.skipTest('qemu.download.dir is not set')
    # We use some known reference
    # FIXME: This is not used by callers, something is wrong -- vila 2019-11-14
    reference_cloud_image_name = 'ubuntu/bionic-amd64'
    cloud_image_path = os.path.join(
        download_cache, reference_cloud_image_name)
    if not os.path.exists(cloud_image_path):
        test.skipTest('{} is not available'.format(cloud_image_path,))
    # /!\ Tests should be careful when using this images_dir shared
    # resource. No existing image should be deleted.
    # I.e. use unique names, cleanup.
    # FIXME: It's probably better to create hard links to existing images ?
    # -- vila 2019-09-11
    images_dir = user_conf.get('qemu.images.dir')
    return download_cache, reference_cloud_image_name, images_dir


@features.requires(features.wget_feature)
class TestDownloadImage(unittest.TestCase):

    def setUp(self):
        # Downloading real isos or images is too long for tests, instead, we
        # fake it by downloading a small but known to exist file: MD5SUMS
        super(TestDownloadImage, self).setUp()
        fixtures.isolate_from_disk(self)
        download_dir = os.path.join(self.uniq_dir, 'downloads')
        os.mkdir(download_dir)
        fixtures.override_logging(self)
        # 'focal' should stay valid for a while...
        # FIXME: 'focal' should be replaced by some option defining the current
        # LTS -- vila-2019-11-08
        # FIXME: Especially given that the "current" LTS is bionic in almost
        # all other tests -- vila 2019-11-16
        self.download_url = (
            'https://cloud-images.ubuntu.com/focal/current/MD5SUMS')
        self.conf = config.VmStack('foo')
        self.conf.store._load_from_string('''
vm.name = foo
vm.release = focal
vm.architecture = amd64
qemu.download.dir = {download_dir}
qemu.download.url = {download_url}
'''.format(download_dir=download_dir, download_url=self.download_url))
        self.conf.set('qemu.image.setup', 'download')
        self.conf.set('qemu.image.teardown', 'download')
        self.conf.set('qemu.image', '{qemu.download.path}')

    def test_download(self):
        vm = qemu.Qemu(self.conf)
        vm.setup_image()
        self.assertTrue(os.path.exists(vm.disk_image_path()))
        # Trying to download again will find the file in the cache
        self.assertFalse('Already at' in self.log_stream.getvalue())
        vm.setup_image()
        self.assertTrue('Already at' in self.log_stream.getvalue())

    def test_download_creates_cache(self):
        download_dir = os.path.join(self.uniq_dir, 'I-dont-exist')
        self.conf.set('qemu.download.dir', download_dir)
        vm = qemu.Qemu(self.conf)
        self.assertFalse(os.path.exists(os.path.dirname(vm.disk_image_path())))
        vm.setup_image()
        self.assertTrue(os.path.exists(vm.disk_image_path()))

    def test_download_unknown_fails(self):
        # Sabotage the valid url
        url = self.conf.get('qemu.download.url')
        self.conf.set('qemu.download.url', url + 'I-dont-exist')
        vm = qemu.Qemu(self.conf)
        self.assertRaises(errors.CommandError, vm.setup_image)


@features.requires(features.qemu_img_feature)
class TestConvertImage(unittest.TestCase):

    def setUp(self):
        super(TestConvertImage, self).setUp()
        fixtures.setup_tests_config(self)
        convert_dir = os.path.join(self.uniq_dir, 'convert')
        self.conf = config.VmStack('foo')
        self.vm = qemu.Qemu(self.conf)
        self.conf.set('vm.name', 'foo')
        self.conf.set('vm.release', 'bionic')
        self.conf.set('vm.architecture', 'amd64')
        self.conf.set('vm.disk_size', '2.5G')  # Limit the size for tests
        self.conf.set('qemu.download.dir', self.user_download_dir)
        self.conf.set('qemu.images.dir', convert_dir)
        self.conf.set('qemu.image.setup', 'convert,resize')
        self.conf.set('qemu.image.teardown', 'convert')

    def test_convert_image(self):
        self.assertFalse(os.path.exists(self.vm.disk_image_path()))
        self.vm.setup_image()
        self.assertTrue(os.path.exists(self.vm.disk_image_path()))

    def test_convert_no_source(self):
        self.conf.set('qemu.download.path', 'I-dont-exist')
        with self.assertRaises(errors.CommandError) as cm:
            self.vm.setup_image()
        self.assertEqual(1, cm.exception.retcode)
        self.assertTrue('I-dont-exist' in cm.exception.err)

    def test_convert_too_small(self):
        # This is a lower bound to the reference image which is unlikely to
        # shrink below that.
        self.conf.set('vm.disk_size', '1G')
        self.assertFalse(os.path.exists(self.vm.disk_image_path()))
        with self.assertRaises(errors.CommandError) as cm:
            self.vm.setup_image()
        self.assertEqual(1, cm.exception.retcode)
        self.assertTrue('does not support resize' in cm.exception.err)
        # The resize failed but the image exists (if only to help set the
        # proper size)
        self.assertTrue(os.path.exists(self.vm.disk_image_path()))


@features.requires(features.qemu_img_feature)
class TestClone(unittest.TestCase):

    def setUp(self):
        super(TestClone, self).setUp()
        fixtures.setup_tests_config(self)
        clone_dir = os.path.join(self.uniq_dir, 'clone')
        # Create a shared config
        conf = config.VmStack(None)
        conf.store._load_from_string('''
vm.release=bionic
vm.architecture = amd64
vm.disk_size = 2.5G # Limit the size for tests
qemu.download.dir = {download_dir}
qemu.images.dir = {images_dir}
[base]
vm.name=base
[foo]
vm.name=foo
vm.backing=base
'''.format(download_dir=self.user_download_dir, images_dir=clone_dir))
        conf.store.save()
        conf.store.unload()
        # Points to the reference image
        bconf = config.VmStack('base')
        bconf.set('vm.published_as', '{qemu.download.path}')
        bconf.store.save()
        bconf.store.unload()
        self.conf = config.VmStack('foo')
        self.vm = qemu.Qemu(self.conf)
        self.conf.set('qemu.image.setup', 'clone,resize')
        self.conf.set('qemu.image.teardown', 'clone')

    def test_clone_image(self):
        self.assertFalse(os.path.exists(self.vm.disk_image_path()))
        self.vm.setup_image()
        self.assertTrue(os.path.exists(self.vm.disk_image_path()))

    def test_clone_no_source(self):
        bconf = config.VmStack('base')
        bconf.set('vm.published_as', 'I-dont-exist')
        with self.assertRaises(errors.CommandError) as cm:
            self.vm.setup_image()
        self.assertEqual(1, cm.exception.retcode)
        self.assertTrue('I-dont-exist' in cm.exception.err)

    def test_clone_too_small(self):
        # This is a lower bound to the reference image which is unlikely to
        # shrink below that.
        self.conf.set('vm.disk_size', '1G')
        self.assertFalse(os.path.exists(self.vm.disk_image_path()))
        with self.assertRaises(errors.CommandError) as cm:
            self.vm.setup_image()
        self.assertEqual(1, cm.exception.retcode)
        self.assertTrue('does not support resize' in cm.exception.err)
        # The resize failed but the image exists (if only to help set the
        # proper size)
        self.assertTrue(os.path.exists(self.vm.disk_image_path()))


@features.requires(features.qemu_img_feature)
class TestSeedImage(unittest.TestCase):

    def setUp(self):
        super(TestSeedImage, self).setUp()
        fixtures.setup_tests_config(self)
        images_dir = os.path.join(self.uniq_dir, 'images')
        self.conf = config.VmStack('foo')
        self.vm = qemu.Qemu(self.conf)
        self.conf.set('vm.class', 'qemu')
        self.conf.set('vm.name', 'foo')
        self.conf.set('vm.release', 'bionic')
        self.conf.set('vm.architecture', 'amd64')
        self.conf.set('qemu.images.dir', images_dir)

    def test_create_seed_image(self):
        self.assertTrue(self.vm._seed_path is None)
        self.vm.create_seed_image()
        self.assertFalse(self.vm._seed_path is None)
        self.assertTrue(os.path.exists(self.vm._seed_path))


class TestQemuMonitor(unittest.TestCase):

    def setUp(self):
        super(TestQemuMonitor, self).setUp()
        fixtures.set_uniq_vm_name(self)
        fixtures.setup_tests_config(self)
        # Create a shared config
        conf = config.VmStack(None)
        conf.store._load_from_string('''
[{name}]
vm.name = {name}
vm.class = qemu
qemu.ssh.localhost.port = 10022 # FIXME: Needs a better way -- vila 2019-09-11
vm.release=bionic
vm.architecture = amd64
'''.format(name=self.vm_name))
        conf.store.save()
        conf.store.unload()

    def test_terminate_qemu(self):
        vm = qemu.Qemu(config.VmStack(self.vm_name))
        vm.spawn_qemu()
        self.addCleanup(vm.terminate_qemu)
        self.assertEqual('RUNNING', vm.state())


class TestSetupWithSeed(unittest.TestCase):

    def setUp(self):
        super(TestSetupWithSeed, self).setUp()
        fixtures.set_uniq_vm_name(self, short=True)
        fixtures.setup_tests_config(self)
        fixtures.override_logging(self)
        self.conf = config.VmStack(self.vm_name)
        features.requires_existing_bridge(self, self.conf.get('qemu.bridge'))
        images_dir = os.path.join(self.uniq_dir, 'images')
        self.conf.set('vm.class', 'qemu')
        self.conf.set('vm.update', 'False')  # Shorten install time
        self.conf.set('vm.cpus', '2')
        self.conf.set('vm.release', 'bionic')
        self.conf.set('vm.architecture', 'amd64')
        self.conf.set('vm.disk_size', '2.5G')  # Limit the size for tests
        self.conf.set('qemu.download.dir', self.user_download_dir)
        self.conf.set('qemu.images.dir', images_dir)
        self.conf.set('qemu.image.setup', 'download,convert,resize')
        self.conf.set('qemu.image.teardown', 'convert,resize')
        self.conf.set('vm.name', self.vm_name)
        self.conf.set('qemu.networks',
                      '-net bridge,br={qemu.bridge}'
                      ' -net nic,macaddr={qemu.mac.address}')
        self.conf.store.save()
        self.conf.store.unload()
        self.vm = qemu.Qemu(self.conf)

    def test_setup_with_seed(self):
        self.addCleanup(self.vm.teardown, force=True)
        self.vm.setup()
        self.assertEqual('RUNNING', self.vm.state())

    def test_start_keeps_mac(self):
        self.addCleanup(self.vm.teardown, force=True)
        self.vm.setup()
        self.assertEqual('RUNNING', self.vm.state())
        mac_address = self.vm.econf.get('qemu.mac.address')
        ip_addresss = self.vm.econf.get('vm.ip')
        self.vm.stop()
        self.vm.start()
        self.assertEquals(mac_address, self.vm.econf.get('qemu.mac.address'))
        # The aim is to get a stable ip so it must be the same on all starts
        self.assertEquals(ip_addresss, self.vm.econf.get('vm.ip'))


class TestSetupWithBacking(unittest.TestCase):

    def setUp(self):
        super(TestSetupWithBacking, self).setUp()
        fixtures.set_uniq_vm_name(self, short=True)
        fixtures.setup_tests_config(self)
        fixtures.override_logging(self)
        # Create a shared config
        conf = config.VmStack(None)
        features.requires_existing_bridge(self, conf.get('qemu.bridge'))
        images_dir = os.path.join(self.uniq_dir, 'images')
        conf.set('vm.class', 'qemu')
        conf.set('vm.update', 'False')  # Shorten install time
        conf.set('vm.release', 'bionic')
        conf.set('vm.architecture', 'amd64')
        conf.set('vm.disk_size', '2.5G')  # Limit the size for tests
        conf.set('qemu.download.dir', self.user_download_dir)
        conf.set('qemu.images.dir', images_dir)

        bconf = config.VmStack('base')
        bconf.set('vm.name', 'base')

        vconf = config.VmStack(self.vm_name)
        vconf.set('vm.name', self.vm_name)
        vconf.set('vm.backing', 'base')

        conf.set('qemu.image.setup', 'clone,resize')
        conf.set('qemu.image.teardown', 'clone')
        conf.set('qemu.networks',
                 '-net bridge,br={qemu.bridge}'
                 ' -net nic,macaddr={qemu.mac.address}')
        # Points to the reference image
        bconf.set('vm.published_as', '{qemu.download.path}')

        # conf, bconf and vconf are all in the same store, a single save is
        # enough
        self.assertTrue(conf.store == vconf.store)
        self.assertTrue(conf.store == bconf.store)
        conf.store.save()
        conf.store.unload()

    def test_setup_with_backing(self):
        vm = qemu.Qemu(config.VmStack(self.vm_name))
        self.addCleanup(vm.teardown, force=True)
        vm.setup()
        self.assertEqual('RUNNING', vm.state())

# FIXME: While vm.update=False makes the test faster, it missed the bug where
# Qemu.setup() wasn't waiting for cloud-init to finish leading to a very
# confusing behavior. Specifically, cloud-init add-apt-repository (for a ppa)
# raced with the first apt-get update issued by byov.
# -- vila 2019-12-03



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals


class BaseError(Exception):
    """Base class for all byov exceptions.

    :cvar fmt: A format string that daughter classes override.
    """

    fmt = 'Daughter classes should redefine this'

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __unicode__(self):
        return self.fmt.format([], **self.__dict__)

    __repr__ = __unicode__
    __str__ = __unicode__


class ByovError(BaseError):

    def __init__(self, fmt, **kwargs):
        super(ByovError, self).__init__(**kwargs)
        self.fmt = fmt


class VmUnknown(BaseError):

    fmt = '"{name}" is unknown.'

    def __init__(self, name):
        super(VmUnknown, self).__init__(name=name)


class InvalidVmClass(BaseError):

    fmt = 'vm.class "{value}" is not valid for "{name}".'

    def __init__(self, name, value):
        super(InvalidVmClass, self).__init__(name=name, value=value)


class VmRunning(BaseError):

    fmt = '"{name}" is running.'

    def __init__(self, name):
        super(VmRunning, self).__init__(name=name)


class VmNotRunning(BaseError):

    fmt = '"{name}" is not running.'

    def __init__(self, name):
        super(VmNotRunning, self).__init__(name=name)


class CommandError(BaseError):

    fmt = '''
  command: {joined_cmd}
  retcode: {retcode}
  stdout: {out}
  stderr: {err}
'''

    def __init__(self, cmd, retcode, out, err):
        super(CommandError, self).__init__(joined_cmd=' '.join(cmd),
                                           retcode=retcode, err=err, out=out)


class ConfigOptionNotFound(BaseError):

    fmt = 'Option "{name}" does not exist.'

    def __init__(self, name):
        super(ConfigOptionNotFound, self).__init__(name=name)


class ConfigValueError(BaseError):

    fmt = 'Bad value "{value}" for option "{name}".'

    def __init__(self, name, value):
        super(ConfigValueError, self).__init__(name=name, value=value)


class ConfigPathNotFound(BaseError):

    fmt = 'No such file: {path} from {name}'

    def __init__(self, path, name):
        super(ConfigPathNotFound, self).__init__(path=path, name=name)


class ConsoleEOFError(BaseError):

    fmt = 'Encountered EOF on console, something went wrong'


class CloudInitError(BaseError):

    fmt = 'cloud-init reported: {line} check your config'

    def __init__(self, line):
        super(CloudInitError, self).__init__(line=line)


class SubIdError(BaseError):

    fmt = 'The /etc/sub{typ} file has no "root:{subid}:1" line'

    def __init__(self, typ, subid):
        super(SubIdError, self).__init__(typ=typ, subid=subid)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import errno
import os


from byov import subprocesses


def infos_from_path(key_path):
    """Analyze path to find ssh key type and kind.

    The basename should begin with the ssh type used to create the key and end
    with '.pub' for a public key.

    If the type is neither of rsa, dsa, ecdsa or ed25519, None is returned.

    :param key_path: A path to an ssh key.

    :return: (type, kind) 'type' is the ssh key type or None if neither of rds,
        dsa or ecdsa. 'kind' is 'public' if the path ends with '.pub',
        'private' otherwise.

    """
    base = os.path.basename(key_path)
    for p in ('rsa', 'dsa', 'ecdsa', 'ed25519'):
        if base.startswith(p):
            ssh_type = p
            break
    else:
        ssh_type = None
    if base.endswith('.pub'):
        kind = 'public'
    else:
        kind = 'private'
    return ssh_type, kind


def keygen(key_path, comment):
    """Generate an ssh keypair.

    :param key_path: The private path where the key should be generated (the
        public key path is obtained by adding the '.pub' suffix.

    :param comment: The comment to embed in the key for documentation and
        identification purposes.
    """

    ssh_type, kind = infos_from_path(key_path)
    if kind == 'private':  # public will be generated at the same time

        def rm_f(p):
            try:
                os.remove(p)
            except OSError as e:  # python2 doesn't provide FileNotFoundError
                # It doesn't exist ? Job done.
                if e.errno != errno.ENOENT:
                    raise
        rm_f(key_path)
        rm_f(os.path.join(key_path, '.pub'))
        # FIXME: Increase key length via '-b' respecting key type constraints
        # (see man ssh-keygen) -- vila 2016-07-18
        subprocesses.run(['ssh-keygen', '-f', key_path, '-N', '',
                          '-t', ssh_type,
                          '-C', comment])



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals
import logging
import subprocess


from byov import errors


logger = logging.getLogger(__name__)


def run(args, cmd_input=None, raise_on_error=True):
    """Run the specified command capturing output and errors.

    :param args: A list of a command and its arguments.

    :param cmd_input: A unicode string to feed the command with.

    :param raise_on_error: A boolean controlling whether or not an exception is
        raised if the command fails.

    :return: A tuple of the return code, the output and the errors as unicode
        strings.

    """
    stdin = None
    if cmd_input is not None:
        stdin = subprocess.PIPE
        cmd_input = cmd_input.encode('utf8')
    logger.debug('Running {}'.format(' '.join(args)))
    proc = subprocess.Popen(args,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            stdin=stdin)
    out, err = proc.communicate(cmd_input)
    out = out.decode('utf8')
    err = err.decode('utf8')
    if raise_on_error and proc.returncode:
        raise errors.CommandError(args, proc.returncode, out, err)
    logger.debug('Returned: {}'.format(proc.returncode))
    if out:
        logger.debug('stdout: {}'.format(out))
    if err:
        logger.debug('stderr: {}'.format(err))
    return proc.returncode, out, err


def pipe(args):
    """Run the specified command as a pipe.

    The caller is responsible for processing the output and the errors.

    :param args:  A list of a command and its arguments.

    :return: The Popen object.
    """
    proc = subprocess.Popen(args,
                            stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    return proc



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals
import errno
import io
import os
import unittest

import byov

from byoc import (
    errors as config_errors,
    options,
    registries,
)
from byot import assertions
from byov import (
    commands,
    config,
    errors,
    vms,
)
from byov.tests import fixtures


class TestHelpOptions(unittest.TestCase):

    def setUp(self):
        super(TestHelpOptions, self).setUp()
        self.out = io.StringIO()
        self.err = io.StringIO()

    def parse_args(self, args):
        help_cmd = commands.Help(out=self.out, err=self.err)
        return help_cmd.parse_args(args)

    def test_defaults(self):
        ns = self.parse_args([])
        self.assertEqual([], ns.options)

    def test_single_option(self):
        ns = self.parse_args(['ssh.key'])
        self.assertEqual(['ssh.key'], ns.options)

    def test_several_options(self):
        ns = self.parse_args(['vm.name', 'ssh.key', 'not-an-option'])
        self.assertEqual(['vm.name', 'ssh.key', 'not-an-option'], ns.options)


class TestHelp(unittest.TestCase):

    def setUp(self):
        super(TestHelp, self).setUp()
        self.out = io.StringIO()
        self.err = io.StringIO()
        self.help_cmd = commands.Help(out=self.out, err=self.err)

    def assertHelp(self, expected, args=None):
        if args is None:
            args = []
        self.help_cmd.parse_args(args)
        self.help_cmd.run()
        assertions.assertMultiLineAlmostEqual(self, expected,
                                              self.out.getvalue())

    def test_help_commands(self):
        # Separate first line and its trailing whitespace (go away pep8)
        self.assertHelp('byov.commands: \n' + '''\
\tconfig: Manage a virtual machine configuration.
\tdigest: Display the digest of a virtual machine.
\thelp: Describe byov configuration options.
\tpublish: Publish a virtual machine image.
\tpull: Pull a remote file from an existing virtual machine.
\tpush: Push local file to an existing virtual machine.
\tsetup: Setup a virtual machine.
\tshell: Start a shell, run a command or a script inside a virtual machine.
\tssh-authorize: Authorize a key in the vm for ssh access.
\tssh-register: Register a host in the vm for ssh access.
\tstart: Start an existing virtual machine.
\tstatus: Display the status of a virtual machine.
\tstop: Stop an existing virtual machine.
\tteardown: Teardown a virtual machine.
\tversion: Output the byov version.

''',
                        ['byov.commands'])

    def test_help_help(self):
        with self.assertRaises(SystemExit):
            self.help_cmd.parse_args(['-h'])
        assertions.assertMultiLineAlmostEqual(self, '''\
usage: byovm... help [-h] [OPTION [OPTION ...]]

Describe byov configuration options.

positional arguments:
  OPTION      Display help for each option (Topics if none is given).

optional arguments:
  -h, --help  show this help message and exit
''',
                                              self.out.getvalue())

    def test_help_regexp(self):
        self.assertHelp('''\
debian.lxd.image: The debian lxd image to boot from.
debian.package.manager: The package manager for debian.
debian.user: The default user for debian.
debian.user.shell: The default shell on debian.
''',
                        ['debian.*'])

    def test_help_not_matching_regexps(self):
        regexps = ['^IDontExist.*', 'IDoNotExistEither$']
        self.assertHelp('', regexps)
        self.assertEqual(
            'No options matched {}\n'.format(regexps),
            self.err.getvalue())


class FakeVM(vms.VM):
    """A fake VM for tests that doesn't trigger dangerous or costly calls."""

    # Must be set by test (and reset at tearDown)
    states = None

    def __init__(self, conf):
        super(FakeVM, self).__init__(conf)
        self.setup_called = False
        self.start_called = False
        self.shell_called = False
        self.shell_command = None
        self.shell_cmd_args = None
        self.publish_called = False
        self.pull_called = False
        self.pull_remote = None
        self.pull_local = None
        self.push_called = False
        self.push_local = None
        self.push_remote = None
        self.stop_called = False
        self.teardown_called = False

    def state(self):
        return self.states.get(self.conf.get('vm.name'), None)

    def setup(self):
        self.setup_called = True

    def start(self):
        self.start_called = True

    def shell(self, command, *cmd_args):
        self.shell_called = True
        self.shell_command = command
        self.shell_cmd_args = cmd_args
        # Respect signature
        return 0, None, None

    def stop(self):
        self.stop_called = True

    def publish(self):
        self.publish_called = True

    def pull(self, remote, local):
        self.pull_called = True
        self.pull_remote = remote
        self.pull_local = local

    def push(self, local, remote):
        self.push_called = True
        self.push_local = local
        self.push_remote = remote

    def teardown(self, force=False):
        self.teardown_called = True


def register_fake_vm(test):
    # Register our fake vm class inside the registry option as that's where it
    # matters.
    vm_reg = options.option_registry.get('vm.class')
    fixtures.patch(test, vm_reg, 'registry', registries.Registry())
    # Only one kind of vm exists in the temporary registry
    vm_reg.registry.register('fake', FakeVM, 'Fake vm')
    FakeVM.states = {}

    def reset_fake_states():
        FakeVM.states = None
    test.addCleanup(reset_fake_states)


def setup_fake_vm_conf(test, name):
    # Define a user config file for the vm
    dir_path = os.path.join(config.user_config_dir(), 'conf.d')
    c_path = os.path.join(dir_path, '{}.conf'.format(name))
    with open(c_path, 'w') as f:
        f.write('''
[{name}]
vm.name={name}
vm.class=fake
vm.release=bionic
vm.architecture=amd64
        '''.format(name=name))
    return config.VmStack(name)


class TestWithVmFoo(unittest.TestCase):

    def setUp(self):
        super(TestWithVmFoo, self).setUp()
        fixtures.isolate_from_disk(self)
        register_fake_vm(self)
        FakeVM.states = {}
        dir_path = os.path.join(config.user_config_dir(), 'conf.d')
        os.makedirs(dir_path)
        self.conf = setup_fake_vm_conf(self, 'foo')
        self.out = io.StringIO()
        self.err = io.StringIO()


class TestVmCommandOptions(TestWithVmFoo):

    def setUp(self):
        super(TestVmCommandOptions, self).setUp()
        self.vm_cmd = commands.VmCommand(out=self.out, err=self.err)

    def parse_args(self, args):
        return self.vm_cmd.parse_args(args)

    def test_nothing(self):
        with self.assertRaises(SystemExit):
            self.parse_args([])

    def test_defaults(self):
        ns = self.parse_args(['foo'])
        self.assertEqual('foo', ns.vm_name)

    def test_overrides(self):
        ns = self.parse_args(['foo', '--option', 'vm.name=bug',
                              '-O', 'foo=bar', '-Obaz=quux'])
        self.assertEqual('foo', ns.vm_name)
        self.assertEqual(['vm.name=bug', 'foo=bar', 'baz=quux', 'vm.name=foo'],
                         ns.overrides)
        # vm.name override has been ignored
        self.assertEqual('foo', self.vm_cmd.vm.conf.get('vm.name'))
        # The other options are available for the vm
        self.assertEqual('bar', self.vm_cmd.vm.conf.get('foo'))
        self.assertEqual('quux', self.vm_cmd.vm.conf.get('baz'))

    def test_error_in_overrides(self):
        with self.assertRaises(SystemExit):
            self.parse_args(['foo', '-O'])
        with self.assertRaises(config_errors.InvalidOverrideError):
            self.parse_args(['foo', '-O', 'toto'])


class TestVmCommand(TestWithVmFoo):

    def run_cmd(self, args):
        vm_cmd = commands.VmCommand()
        vm_cmd.parse_args(args)
        vm_cmd.run()

    def test_unknown(self):
        self.conf.set('vm.name', 'I-dont-exist')
        with self.assertRaises(errors.ByovError):
            self.run_cmd(['I-dont-exist'])

    def test_unknown_vm_class(self):
        self.conf.set('vm.class', 'I-dont-exist')
        with self.assertRaises(errors.InvalidVmClass):
            self.run_cmd(['foo'])


class TestConfigOptions(TestWithVmFoo):

    def parse_args(self, args):
        conf_cmd = commands.Config(out=self.out, err=self.err)
        return conf_cmd.parse_args(args)

    def test_nothing(self):
        with self.assertRaises(SystemExit):
            self.parse_args([])

    def test_defaults(self):
        ns = self.parse_args(['foo'])
        self.assertEqual('foo', ns.vm_name)
        self.assertIs(None, ns.name)
        self.assertFalse(ns.all)
        self.assertFalse(ns.remove)

    def test_name(self):
        ns = self.parse_args(['foo', 'bar'])
        self.assertEqual('bar', ns.name)

    def assertError(self, expected):
        assertions.assertMultiLineAlmostEqual(self, expected,
                                              self.err.getvalue())

    def test_too_much_args(self):
        with self.assertRaises(SystemExit) as cm:
            self.parse_args(['foo', 'bar', 'baz'])
        self.assertEqual(2, cm.exception.code)
        self.assertTrue(self.err.getvalue().endswith(
            'config: error: unrecognized arguments: baz\n'))

    def test_remove_requires_name(self):
        with self.assertRaises(errors.ByovError) as cm:
            self.parse_args(['foo', '--remove'])
        self.assertEqual('--remove expects an option to remove.',
                         str(cm.exception))

    def test_remove_all_fails(self):
        with self.assertRaises(errors.ByovError) as cm:
            self.parse_args(['foo', '--remove', '--all'])
        self.assertEqual('--remove and --all are mutually exclusive.',
                         str(cm.exception))

    def test_set_only_one_option(self):
        with self.assertRaises(errors.ByovError) as cm:
            self.parse_args(['foo', 'bar=baz', '--all'])
        self.assertEqual('Only one option can be set.',
                         str(cm.exception))


class TestConfig(TestWithVmFoo):

    def run_config(self, args):
        self.vm = FakeVM(self.conf)
        cmd_config = commands.Config(_vm=self.vm, out=self.out, err=self.err)
        cmd_config.parse_args(args)
        cmd_config.run()

    def test_set_new_option(self):
        self.assertEqual(None, self.conf.get('foo'))
        self.run_config(['foo', 'foo=bar'])
        self.assertEqual('bar', self.conf.get('foo'))

    def test_set_existing_option(self):
        self.conf.set('foo', 'bar')
        self.assertEqual('bar', self.conf.get('foo'))
        self.run_config(['foo', 'foo=qux'])
        self.assertEqual('qux', self.conf.get('foo'))

    def test_remove_unknown_option(self):
        with self.assertRaises(errors.ConfigOptionNotFound):
            self.run_config(['foo', '-r', 'foo'])

    def test_remove_existing_option(self):
        self.conf.set('bar', 'baz')
        self.run_config(['foo', '-r', 'bar'])
        self.assertEqual(None, self.conf.get('bar'))

    def assertOutput(self, expected):
        assertions.assertMultiLineAlmostEqual(self, expected,
                                              self.out.getvalue())

    def test_list_option(self):
        self.conf.set('foo', 'bar')
        self.run_config(['foo', 'foo'])
        self.assertOutput('bar')

    def test_unknown_option(self):
        with self.assertRaises(errors.ConfigOptionNotFound):
            self.run_config(['foo', 'foo'])
        self.assertOutput('')

    def test_list_that_conf(self):
        self.run_config(['foo'])
        self.assertOutput('''\
.../home/.config/byov/conf.d/foo.conf:
  [foo]
  vm.name = foo
  vm.class = fake
  vm.release = bionic
  vm.architecture = amd64
''')

    def test_more_options(self):
        self.conf.set('one', '1')
        self.conf.set('two', '2')
        self.run_config(['foo'])
        self.assertOutput('''\
.../home/.config/byov/byov.conf:
  [foo]
  one = 1
  two = 2
.../home/.config/byov/conf.d/foo.conf:
  [foo]
  vm.name = foo
  vm.class = fake
  vm.release = bionic
  vm.architecture = amd64
''')

    def test_options_several_sections(self):
        self.conf.store.unload()
        self.conf.store._load_from_string('''\
one = 1
two = 2
[foo]
one = 1
two = foo
[bar]
one = bar
two = foo
''')
        self.run_config(['foo'])
        self.assertOutput('''\
.../home/.config/byov/byov.conf:
  [foo]
  one = 1
  two = foo
.../home/.config/byov/conf.d/foo.conf:
  [foo]
  vm.name = foo
  vm.class = fake
  vm.release = bionic
  vm.architecture = amd64
.../home/.config/byov/byov.conf:
  []
  one = 1
  two = 2
''')

    def test_home_config_shown_twice(self):
        self.conf.store.unload()
        self.conf.store._load_from_string('''\
one = 1
[foo]
one = foo
[bar]
one = bar
''')
        self.conf.store.save()
        # Reload the stack from the current dir
        os.chdir('home/.config/byov')
        self.conf = config.VmStack('foo')
        self.run_config(['foo'])
        # Going under ~/.config/byov is weird enough so the user seemingly get
        # duplicates (but only the first match is used anyway).
        self.assertOutput('''\
.../home/.config/byov/byov.conf:
  [foo]
  one = foo
  one = foo
.../home/.config/byov/conf.d/foo.conf:
  [foo]
  vm.name = foo
  vm.class = fake
  vm.release = bionic
  vm.architecture = amd64
.../home/.config/byov/byov.conf:
  []
  one = 1
  one = 1
''')

    def test_file_not_found(self):
        with self.assertRaises(errors.ConfigPathNotFound):
            self.run_config(['foo', '@I-dont-exist'])

    def test_file_found(self):
        fname = 'bar'
        with open(fname, 'w') as f:
            f.write('{vm.name}\n')
        self.run_config(['foo', '@bar'])
        self.assertOutput('foo\n')


class TestSetupOptions(TestWithVmFoo):

    def parse_args(self, args):
        self.vm = FakeVM(self.conf)
        setup = commands.Setup(_vm=self.vm, out=self.out, err=self.err)
        return setup.parse_args(args)

    def test_nothing(self):
        with self.assertRaises(SystemExit):
            self.parse_args([])

    def test_defaults(self):
        ns = self.parse_args(['foo'])
        self.assertEqual('foo', ns.vm_name)
        self.assertFalse(ns.download)
        self.assertFalse(ns.ssh_keygen)

    def test_download(self):
        ns = self.parse_args(['foo', '--download'])
        self.assertEqual('foo', ns.vm_name)
        self.assertTrue(ns.download)
        self.assertFalse(ns.ssh_keygen)

    def test_ssh_keygen(self):
        ns = self.parse_args(['foo', '--ssh-keygen'])
        self.assertEqual('foo', ns.vm_name)
        self.assertFalse(ns.download)
        self.assertTrue(ns.ssh_keygen)


class TestSetup(TestWithVmFoo):

    def run_setup(self, args, vm=None):
        if vm is None:
            self.vm = FakeVM(self.conf)
        setup = commands.Setup(_vm=self.vm)
        setup.parse_args(args)
        setup.run()

    def test_not_running(self):
        FakeVM.states = {'foo': 'shut off'}
        self.run_setup(['foo'])
        self.assertTrue(self.vm.setup_called)
        self.assertTrue(self.vm.teardown_called)

    def test_while_running(self):
        FakeVM.states = {'foo': 'running'}
        with self.assertRaises(errors.VmRunning):
            self.run_setup(['foo'])
        self.assertFalse(self.vm.setup_called)
        self.assertFalse(self.vm.teardown_called)

    def test_force_while_running(self):
        FakeVM.states = {'foo': 'running'}
        self.vm = FakeVM(self.conf)

        # Override stop so we can change the state when it's called
        def stop():
            FakeVM.stop(self.vm)
            # setup() will query state again
            FakeVM.states['foo'] = 'shut off'
        self.vm.stop = stop
        self.run_setup(['--force', 'foo'], vm=self.vm)
        self.assertFalse(self.vm.stop_called)
        self.assertTrue(self.vm.teardown_called)
        self.assertTrue(self.vm.setup_called)


class TestDigest(TestWithVmFoo):

    def run_digest(self, args):
        self.vm = FakeVM(self.conf)
        cmd_digest = commands.Digest(_vm=self.vm, out=self.out, err=self.err)
        cmd_digest.parse_args(args)
        cmd_digest.run()

    def assertOutput(self, expected):
        assertions.assertMultiLineAlmostEqual(self, expected,
                                              self.out.getvalue())

    def test_unknown(self):
        cmd_digest = commands.Digest(out=self.out, err=self.err)
        cmd_digest.parse_args(['I-dont-exist'])
        cmd_digest.run()
        self.assertOutput('UNKNOWN')


class TestStatus(TestWithVmFoo):

    def run_status(self, args):
        self.vm = FakeVM(self.conf)
        status = commands.Status(_vm=self.vm, out=self.out, err=self.err)
        status.parse_args(args)
        status.run()

    def test_not_setup(self):
        FakeVM.states = {'foo': 'UNKNOWN'}
        self.run_status(['foo'])
        self.assertEqual('UNKNOWN\n', self.out.getvalue())

    def test_not_running(self):
        FakeVM.states = {'foo': 'shut off'}
        self.run_status(['foo'])
        self.assertEqual('STOPPED\n', self.out.getvalue())

    def test_running(self):
        FakeVM.states = {'foo': 'running'}
        self.run_status(['foo'])
        self.assertEqual('RUNNING\n', self.out.getvalue())


class TestStart(TestWithVmFoo):

    def run_start(self, args):
        self.vm = FakeVM(self.conf)
        start = commands.Start(_vm=self.vm)
        start.parse_args(args)
        start.run()

    def test_not_running(self):
        FakeVM.states = {'foo': 'shut off'}
        self.run_start(['foo'])
        self.assertTrue(self.vm.start_called)

    def test_while_running(self):
        FakeVM.states = {'foo': 'running'}
        with self.assertRaises(errors.VmRunning):
            self.run_start(['foo'])
        self.assertFalse(self.vm.start_called)

    def test_unknown(self):
        self.conf.set('vm.name', 'I-dont-exist')
        with self.assertRaises(errors.VmUnknown):
            self.run_start(['I-dont-exist'])
        self.assertFalse(self.vm.start_called)


class TestShellOptions(TestWithVmFoo):

    def parse_args(self, args):
        self.vm = FakeVM(self.conf)
        setup = commands.Shell(_vm=self.vm, out=self.out, err=self.err)
        return setup.parse_args(args)

    def test_defaults(self):
        ns = self.parse_args(['foo'])
        self.assertEqual('foo', ns.vm_name)
        self.assertIsNone(ns.command)
        self.assertEqual([], ns.args)

    def test_command_without_arguments(self):
        ns = self.parse_args(['foo', 'doit'])
        self.assertEqual('foo', ns.vm_name)
        self.assertEqual('doit', ns.command)
        self.assertEqual([], ns.args)

    def test_command_with_arguments(self):
        ns = self.parse_args(['foo', 'doit', '-a', 'b', 'c'])
        self.assertEqual('foo', ns.vm_name)
        self.assertEqual('doit', ns.command)
        self.assertEqual(['-a', 'b', 'c'], ns.args)


class TestShell(TestWithVmFoo):

    def run_shell(self, args):
        self.vm = FakeVM(self.conf)
        shell = commands.Shell(_vm=self.vm)
        shell.parse_args(args)
        shell.run()

    def test_not_running(self):
        FakeVM.states = {'foo': 'shut off'}
        with self.assertRaises(errors.VmNotRunning):
            self.run_shell(['foo'])
        self.assertFalse(self.vm.shell_called)
        self.assertIsNone(self.vm.shell_command)
        self.assertIsNone(self.vm.shell_cmd_args)

    def test_while_running_no_command(self):
        FakeVM.states = {'foo': 'running'}
        self.run_shell(['foo'])
        self.assertTrue(self.vm.shell_called)
        self.assertIsNone(self.vm.shell_command)
        self.assertEqual((), self.vm.shell_cmd_args)

    def test_while_running_with_command(self):
        FakeVM.states = {'foo': 'running'}
        self.run_shell(['foo', 'doit', 'a', 'b', 'c'])
        self.assertTrue(self.vm.shell_called)
        self.assertEqual('doit', self.vm.shell_command)
        self.assertEqual(('a', 'b', 'c'), self.vm.shell_cmd_args)

    def test_unknown(self):
        self.conf.set('vm.name', 'I-dont-exist')
        with self.assertRaises(errors.VmUnknown):
            self.run_shell(['I-dont-exist'])
        self.assertFalse(self.vm.shell_called)


class TestPullOptions(TestWithVmFoo):

    def parse_args(self, args):
        self.vm = FakeVM(self.conf)
        setup = commands.Pull(_vm=self.vm, out=self.out, err=self.err)
        return setup.parse_args(args)

    def test_nothing(self):
        with self.assertRaises(SystemExit):
            self.parse_args([])

    def test_defaults(self):
        ns = self.parse_args(['foo', 'remote', 'local'])
        self.assertEqual('foo', ns.vm_name)
        self.assertEqual('remote', ns.remote)
        self.assertEqual('local', ns.local)


class TestPull(TestWithVmFoo):

    def run_pull(self, args):
        self.vm = FakeVM(self.conf)
        pull = commands.Pull(_vm=self.vm)
        pull.parse_args(args)
        pull.run()

    def test_not_running(self):
        FakeVM.states = {'foo': 'shut off'}
        with self.assertRaises(errors.VmNotRunning):
            self.run_pull(['foo', 'remote', 'local'])
        self.assertFalse(self.vm.pull_called)
        self.assertEqual(None, self.vm.pull_remote)
        self.assertEqual(None, self.vm.pull_local)

    def test_while_running(self):
        FakeVM.states = {'foo': 'running'}
        self.run_pull(['foo', 'remote', 'local'])
        self.assertTrue(self.vm.pull_called)
        self.assertEqual('remote', self.vm.pull_remote)
        self.assertEqual('local', self.vm.pull_local)

    def test_unknown(self):
        self.conf.set('vm.name', 'I-dont-exist')
        with self.assertRaises(errors.VmUnknown):
            self.run_pull(['I-dont-exist', 'local', 'remote'])
        self.assertFalse(self.vm.pull_called)
        self.assertEqual(None, self.vm.pull_remote)
        self.assertEqual(None, self.vm.pull_local)


class TestPushOptions(TestWithVmFoo):

    def parse_args(self, args):
        self.vm = FakeVM(self.conf)
        setup = commands.Push(_vm=self.vm, out=self.out, err=self.err)
        return setup.parse_args(args)

    def test_nothing(self):
        with self.assertRaises(SystemExit):
            self.parse_args([])

    def test_defaults(self):
        ns = self.parse_args(['foo', 'local', 'remote'])
        self.assertEqual('foo', ns.vm_name)
        self.assertEqual('local', ns.local)
        self.assertEqual('remote', ns.remote)


class TestPush(TestWithVmFoo):

    def run_push(self, args):
        self.vm = FakeVM(self.conf)
        push = commands.Push(_vm=self.vm)
        push.parse_args(args)
        push.run()

    def test_not_running(self):
        FakeVM.states = {'foo': 'shut off'}
        with self.assertRaises(errors.VmNotRunning):
            self.run_push(['foo', 'local', 'remote'])
        self.assertFalse(self.vm.push_called)
        self.assertEqual(None, self.vm.push_local)
        self.assertEqual(None, self.vm.push_remote)

    def test_while_running(self):
        FakeVM.states = {'foo': 'running'}
        self.run_push(['foo', 'local', 'remote'])
        self.assertTrue(self.vm.push_called)
        self.assertEqual('local', self.vm.push_local)
        self.assertEqual('remote', self.vm.push_remote)

    def test_unknown(self):
        self.conf.set('vm.name', 'I-dont-exist')
        with self.assertRaises(errors.VmUnknown):
            self.run_push(['I-dont-exist', 'local', 'remote'])
        self.assertFalse(self.vm.push_called)
        self.assertEqual(None, self.vm.push_local)
        self.assertEqual(None, self.vm.push_remote)


class TestStop(TestWithVmFoo):

    def run_stop(self):
        self.vm = FakeVM(self.conf)
        stop = commands.Stop(_vm=self.vm)
        stop.parse_args(['foo'])
        stop.run()

    def test_not_running(self):
        FakeVM.states = {'foo': 'shut off'}
        with self.assertRaises(errors.VmNotRunning):
            self.run_stop()
        self.assertFalse(self.vm.stop_called)

    def test_while_running(self):
        FakeVM.states = {'foo': 'running'}
        self.run_stop()
        self.assertTrue(self.vm.stop_called)

    def test_unknown(self):
        self.conf.set('vm.name', 'I-dont-exist')
        FakeVM.states = {}
        with self.assertRaises(errors.VmUnknown):
            self.run_stop()
        self.assertFalse(self.vm.stop_called)


class TestPublish(TestWithVmFoo):

    def run_publish(self, args):
        self.vm = FakeVM(self.conf)
        Publish = commands.Publish(_vm=self.vm)
        Publish.parse_args(args)
        Publish.run()

    def test_not_running(self):
        FakeVM.states = {'foo': 'shut off'}
        self.run_publish(['foo'])
        self.assertTrue(self.vm.publish_called)

    def test_while_running(self):
        FakeVM.states = {'foo': 'running'}
        with self.assertRaises(errors.VmRunning):
            self.run_publish(['foo'])
        self.assertFalse(self.vm.publish_called)

    def test_unknown(self):
        self.conf.set('vm.name', 'I-dont-exist')
        FakeVM.states = {}
        with self.assertRaises(errors.VmUnknown):
            self.run_publish(['I-dont-exist'])
        self.assertFalse(self.vm.publish_called)


class TestTeardown(TestWithVmFoo):

    def run_teardown(self, args):
        self.vm = FakeVM(self.conf)
        teardown = commands.Teardown(_vm=self.vm)
        teardown.parse_args(args)
        teardown.run()

    def test_not_running(self):
        FakeVM.states = {'foo': 'shut off'}
        self.run_teardown(['foo'])
        self.assertTrue(self.vm.teardown_called)

    def test_while_running(self):
        FakeVM.states = {'foo': 'running'}
        with self.assertRaises(errors.VmRunning):
            self.run_teardown(['foo'])
        self.assertFalse(self.vm.teardown_called)

    def test_force_while_running(self):
        FakeVM.states = {'foo': 'running'}
        self.run_teardown(['foo', '--force'])
        self.assertTrue(self.vm.teardown_called)

    def test_unknown(self):
        self.conf.set('vm.name', 'I-dont-exist')
        FakeVM.states = {}
        with self.assertRaises(errors.VmUnknown):
            self.run_teardown(['I-dont-exist'])
        self.assertFalse(self.vm.teardown_called)


class TestVersionFunction(unittest.TestCase):

    def test_dev(self):
        self.assertEqual('0.0.1dev8', byov.version((0, 0, 1, 'dev', 8)))

    def test_final(self):
        self.assertEqual('0.0.1', byov.version((0, 0, 1, 'final', 8)))


class TestVersion(unittest.TestCase):

    def setUp(self):
        super(TestVersion, self).setUp()
        self.out = io.StringIO()
        self.err = io.StringIO()
        self.version_cmd = commands.Version(out=self.out, err=self.err)

    def assertVersion(self, expected, args=None):
        if args is None:
            args = []
        self.version_cmd.parse_args(args)
        self.version_cmd.run()
        self.assertEqual(expected, self.out.getvalue())
        self.assertEqual('', self.err.getvalue())

    def test_dev(self):
        fixtures.patch(self, byov, '__version__', (1, 2, 3, 'dev', 4))
        self.assertVersion('1.2.3dev4\n')

    def test_final(self):
        fixtures.patch(self, byov, '__version__', (1, 2, 3, 'final', 4))
        self.assertVersion('1.2.3\n')


class TestSshRegisterOptions(TestWithVmFoo):

    def parse_args(self, args):
        self.vm = FakeVM(self.conf)
        register = commands.SshRegister(
            _vm=self.vm, out=self.out, err=self.err)
        return register.parse_args(args)

    def test_nothing(self):
        with self.assertRaises(SystemExit):
            self.parse_args([])

    def test_defaults(self):
        ns = self.parse_args(['foo', 'target'])
        self.assertEqual('foo', ns.vm_name)
        self.assertEqual('rsa', ns.type)
        self.assertEqual('target', ns.host)


class TestSshRegisterErrors(TestWithVmFoo):

    def setUp(self):
        super(TestSshRegisterErrors, self).setUp()
        self.host_conf = setup_fake_vm_conf(self, 'host')

    def run_ssh_register(self, args):
        self.vm = FakeVM(self.conf)
        self.host = FakeVM(self.host_conf)
        ssh_register = commands.SshRegister(
            _vm=self.vm, out=self.out, err=self.err)
        ssh_register.parse_args(args)
        ssh_register.run()

    def test_unknown(self):
        with self.assertRaises(errors.VmUnknown):
            self.run_ssh_register(['I-dont-exist', 'host'])

    def test_not_running(self):
        FakeVM.states = {'foo': 'shut off', 'host': 'shut off'}
        with self.assertRaises(errors.VmNotRunning):
            self.run_ssh_register(['foo', 'host'])

    def test_host_unknown(self):
        FakeVM.states = {'foo': 'running'}
        with self.assertRaises(errors.VmUnknown):
            self.run_ssh_register(['foo', 'I-dont-exist'])

    def test_host_not_running(self):
        FakeVM.states = {'foo': 'running', 'host': 'shut off'}
        with self.assertRaises(errors.VmNotRunning):
            self.run_ssh_register(['foo', 'host'])


class TestSshAuthorizeOptions(TestWithVmFoo):

    def parse_args(self, args):
        self.vm = FakeVM(self.conf)
        register = commands.SshAuthorize(
            _vm=self.vm, out=self.out, err=self.err)
        return register.parse_args(args)

    def test_nothing(self):
        with self.assertRaises(SystemExit):
            self.parse_args([])

    def test_defaults(self):
        ns = self.parse_args(['foo', 'key'])
        self.assertEqual('foo', ns.vm_name)
        self.assertEqual('key', ns.key)


class TestSshAuthorizeErrors(TestWithVmFoo):

    def run_ssh_authorize(self, args):
        self.vm = FakeVM(self.conf)
        ssh_authorize = commands.SshAuthorize(
            _vm=self.vm, out=self.out, err=self.err)
        ssh_authorize.parse_args(args)
        ssh_authorize.run()

    def test_unknown_vm(self):
        with self.assertRaises(errors.VmUnknown):
            self.run_ssh_authorize(['I-dont-exist', 'not-used'])

    def test_unknown_key(self):
        FakeVM.states = {'foo': 'running'}
        with self.assertRaises(IOError) as cm:
            self.run_ssh_authorize(['foo', 'I-dont-exist'])
        self.assertEqual(errno.ENOENT, cm.exception.errno)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals
import functools
import io
import logging
import unittest


from byot import (
    assertions,
    results,
)

# Since config.vm_class_registry is populated when commands is imported and
# since some tests are parametrized based on its content, it should be imported
# early.
from byov import commands
# Make sure pep8 is happy
__fake = commands.logger
del __fake


class TestLogger(logging.Logger):
    """A logger dedicated to a given test.

    Log messages are captured in a string buffer.
    """

    def __init__(self, test, level=logging.DEBUG,
                 fmt='%(asctime)-15s %(message)s'):
        super(TestLogger, self).__init__(test.id(), level)
        self.stream = io.StringIO()
        handler = logging.StreamHandler(self.stream)
        handler.setFormatter(logging.Formatter(fmt))
        self.addHandler(handler)

    def getvalue(self):
        return self.stream.getvalue()


class log_on_failure(object):
    """Decorates a test to display log on failure.

    This adds a 'logger' attribute to the parameters of the decorated
    test. Using this logger the test can display its content when it fails or
    errors but stay silent otherwise.
    """

    def __init__(self, level=logging.INFO, fmt='%(message)s'):
        self.level = level
        self.fmt = fmt

    def __call__(self, func):

        @functools.wraps(func)
        def decorator(*args):
            test = args[0]
            logger = TestLogger(test, level=self.level, fmt=self.fmt)
            display_log = True

            # We need to delay the log acquisition until we attempt to display
            # it (or we get no content).
            def delayed_display_log():
                msg = 'Failed test log: >>>\n{}\n<<<'.format(logger.getvalue())
                if display_log:
                    raise Exception(msg)

            test.addCleanup(delayed_display_log)

            # Run the test without the decoration
            func(*args + (logger,))
            # If it terminates properly, disable log display
            display_log = False

        return decorator


class TestLogOnFailure(unittest.TestCase):

    def setUp(self):
        super(TestLogOnFailure, self).setUp()
        self.result = results.TextResult(io.StringIO(), verbosity=2)

        # We don't care about timing here so we always return 0 which
        # simplifies matching the expected result
        def zero(atime):
            return 0.0

        self.result.convert_delta_to_float = zero
        # Inner tests will set this from the logger they receive so outter
        # tests can assert the content.
        self.logger = None

    def test_log_not_displayed(self):

        class Test(unittest.TestCase):

            @log_on_failure()
            def test_pass(inner, logger):
                self.logger = logger
                logger.info('pass')

        t = Test('test_pass')
        t.run(self.result)
        self.assertEqual('pass\n', self.logger.getvalue())
        self.assertEqual('{} ... OK (0.000 secs)\n'.format(t.id()),
                         self.result.stream.getvalue())
        self.assertEqual([], self.result.errors)

    def test_log_displayed(self):

        class Test(unittest.TestCase):

            @log_on_failure()
            def test_fail(inner, logger):
                self.logger = logger
                logger.info("I'm broken")
                inner.fail()

        t = Test('test_fail')
        t.run(self.result)
        self.assertEqual("I'm broken\n", self.logger.getvalue())
        # FAILERROR: The test FAIL, the cleanup ERRORs out.
        self.assertEqual(
            '{} ... FAILERROR (0.000 secs)\n'.format(t.id()),
            self.result.stream.getvalue())
        assertions.assertLength(self, 1, self.result.errors)
        failing_test, traceback = self.result.errors[0]
        self.assertIs(t, failing_test)
        expected = traceback.endswith("Failed test log:"
                                      " >>>\nI'm broken\n\n<<<\n")
        self.assertTrue(expected, 'Actual traceback: {}'.format(traceback))

    def test_log_debug_not_displayed(self):

        class Test(unittest.TestCase):

            @log_on_failure()
            def test_debug_silent(inner, logger):
                self.logger = logger
                logger.debug('more info')
                self.fail()

        t = Test('test_debug_silent')
        t.run(self.result)
        self.assertEqual('', self.logger.getvalue())

    def test_log_debug_displayed(self):

        class Test(unittest.TestCase):

            @log_on_failure(level=logging.DEBUG)
            def test_debug_verbose(inner, logger):
                self.logger = logger
                logger.debug('more info')
                self.fail()

        t = Test('test_debug_verbose')
        t.run(self.result)
        self.assertEqual('more info\n', self.logger.getvalue())



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import errno
import unittest

from byot import fixtures
from byov import (
    errors,
    subprocesses,
)


class TestRun(unittest.TestCase):

    def setUp(self):
        super(TestRun, self).setUp()
        fixtures.set_uniq_cwd(self)

    def test_success(self):
        ret, out, err = subprocesses.run(['echo', '-n', 'Hello'])
        self.assertEqual(0, ret)
        self.assertEqual('Hello', out)
        self.assertEqual('', err)

    def test_with_input(self):
        ret, out, err = subprocesses.run(['cat', '-'], 'bar\n')
        self.assertEqual(0, ret)
        self.assertEqual('bar\n', out)
        self.assertEqual('', err)

    def test_failure(self):
        with self.assertRaises(errors.CommandError) as cm:
            subprocesses.run(['ls', 'I-dont-exist'])
        self.assertEqual(2, cm.exception.retcode)
        self.assertEqual('', cm.exception.out)
        self.assertTrue('I-dont-exist' in cm.exception.err)

    def test_raise_on_error(self):
        retcode, out, err = subprocesses.run(
            ['ls', 'I-dont-exist'], raise_on_error=False)
        self.assertEqual(2, retcode)
        self.assertEqual('', out)
        self.assertTrue('I-dont-exist' in err)

    def test_error(self):
        # python2 doesn't have FileNotFoundError
        with self.assertRaises(OSError) as cm:
            subprocesses.run(['I-dont-exist'])
        self.assertEqual(errno.ENOENT, cm.exception.errno)
        self.assertEqual("No such file or directory", cm.exception.strerror)


class TestPipe(unittest.TestCase):

    def setUp(self):
        super(TestPipe, self).setUp()
        fixtures.set_uniq_cwd(self)

    def test_success(self):
        proc = subprocesses.pipe(['echo', '-n', 'Hello'])
        self.assertEqual('Hello', proc.stdout.read().decode('utf8'))

    def test_failure(self):
        proc = subprocesses.pipe(['ls', 'I-dont-exist'])
        self.assertTrue('I-dont-exist' in proc.stdout.read().decode('utf8'))
        # We get the above in stdout because stderr is redirected there
        self.assertIs(None, proc.stderr)

    def test_error(self):
        # python2 doesn't have FileNotFoundError
        with self.assertRaises(OSError) as cm:
            subprocesses.pipe(['I-dont-exist'])
        self.assertEqual(errno.ENOENT, cm.exception.errno)
        self.assertEqual("No such file or directory", cm.exception.strerror)


# MISSINGTESTS:
# test_ssh_no_args (hard, interactive session)
# test_ssh_simple_command
# test_ssh_cant_connect (wrong host, unknown host, missing key, wrong user)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014, 2015, 2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
import errno
import os

from byov import (
    errors,
    subprocesses,
)


class ConsoleMonitor(object):
    """Monitor a console to identify known events."""

    def __init__(self, stream):
        super(ConsoleMonitor, self).__init__()
        self.stream = stream

    def scan(self):
        # FIXME: An overall timeout is needed there (in case things get really
        # wrong with cloud-init) -- vila 2015-04-15
        while True:
            line = self.stream.readline().decode(errors='replace')
            yield line
            if not line:
                raise errors.ConsoleEOFError()
            # MISSINGTEST: ' reboot: Power down\r\n' keeping an eye on finding
            # a more transparent way to test beginning and end of lines
            # (possibly using python universal newlines.  -- vila 2015-05-10
            elif (line.startswith(' * Will now halt') or
                  line.endswith(' reboot: Power down\r\n') or
                  # end of cloud-init with systemd
                  'Stopping Initial cloud-init' in line or
                  # end of wily cloud init with systemd
                  'Stopping Execute cloud user/final scripts' in line or
                  # end of xenial cloud init with systemd
                  'Stopped Execute cloud user/final scripts' in line or
                  "running 'modules:final'" in line):
                # That's our final_message, we're done
                return
            elif ('Failed loading yaml blob' in line or
                  'Unhandled non-multipart userdata starting' in line or
                  'failed to render string to stdout:' in line or
                  'Failed loading of cloud config' in line or
                  # MISSINGTESST: This is cloud-init being more explicit when
                  # it fails because of bugs in user-data -- vila 2015-05-10
                  'Failed running /var/lib/cloud/instance/scripts' in line):
                raise errors.CloudInitError(line)


def actual_file_size(path):
    """Return file size or None if the file doesn't exist.

    :param path: The file of interest.

    :return: 'path' size or None if the file doesn't exist.
    """
    try:
        stat = os.stat(path)
        return stat.st_size
    except OSError as e:
        if e.errno == errno.ENOENT:
            return None
        else:
            raise


class TailMonitor(ConsoleMonitor):

    def __init__(self, path, offset=None):
        cmd = ['tail', '-F', path]
        # MISSINGTEST
        if offset is not None:
            cmd += ['--bytes', '+{}'.format(offset)]
        proc = subprocesses.pipe(cmd)
        super(TailMonitor, self).__init__(proc.stdout)
        self.path = path
        self.cmd = cmd
        self.proc = proc
        self.lines = []

    def scan(self):
        try:
            for line in super(TailMonitor, self).scan():
                # FIXME: Arguably we should decode line from an utf8 encoding
                # as subprocesses.pipe() merges stderr into stdout and utf8
                # error messages have been observed in real life: 'tail: cannot
                # open \xe2\x80\x98/home/vila/vms/lxc1/console\xe2\x80\x99 for
                # reading: Permission denied\n' -- vila 2014-01-19
                self.lines.append(line)
                yield line
        finally:
            self.proc.terminate()
